/*
  2D game engine
  by Matthew Smudz
   Currently written for SFML applications
*/

/*
  DOCUMENTATION

  This header is the keystone of the game!
  Top-level structure:
  engine.h
    game.h
      main.cpp

  The engine contains the functions needed to run the game, but no implementation-specific functions.
  Implementation functions (initializating, updating, and displaying levels) are written in the Game
  class, which is a derived class and therefore inherits all of the engine functions.

  This allows us to keep the implementation separate from the core engine functions, which should remain the same
  in any game.

  A NOTE ON VIEWS

  sf::View is used like a camera in SFML.  Several views with different scrolling speeds can be used to incorporate
  a parallax scrolling effect.  One will often want to compare the position of something in a normal-scrolling foreground
  with something in a slower-scrolling background.  To recover an object's "true" position, subtract the LEFT SIDE position
  of the view you want to translate its position to times the parallax speed of your view.  That is:
  trueX = x - (targetView.GetRect().Left * MY_PARALLAX_SPEED)

  To place an object on a normal-scrolling (parallax value 1.0) view such that it appears at point backgroundPoint
  in a slower-scrolling view, use the following formula:

  desiredX = backgroundPoint * (1.0 / BACKGROUND_PARALLAX) - getScreenWidth()
*/

// Headers
#include <SFML/Graphics.hpp>
#include <SFML/Audio.hpp>
#include <SFML/Network.hpp>
#include <fstream>
#include <sstream>
#include <vector>
#include <utility>

// An animation that can be attached to a sprite or other object.
// Used to store information about that particular animation and execute the animation itself.
// It is expected that all frames in an animation are of equal width (though a derived class could work around that).
class Animation
{
        protected:
                // Used to stagger animation speed
                sf::Clock delayTimer;

                // Current absolute frame (as if all the frames were in linear order)
                int currentFrame;

                // Rows and columns in the sprite sheet
                int rows, columns;

                // Width and height of a given frame
                int frameWidth, frameHeight;

                // Frame to start at
                int firstFrame;

                // Last frame to display
                int lastFrame;

                // Defines the direction of animation (i.e., moving to frames left or right)
                // Restricted to -1 for moving left or +1 for moving right.
                int direction;

        public:
                Animation();

                // Delay between frames (in milliseconds)
                float frameDelay;

                // Should the animation bounce back and forth, or simple loop through frames?
                bool bounce;

                // A derived class can override this if the animation has any peculiarities.
                // NOTE: the default animation just scrolls horizontally across the sprite sheet and loops around to the next row,
                // repeating forever.
                virtual void animate(sf::Sprite& sprite);

                void setRows(int r, const sf::Image *im);
                void setColumns(int c, const sf::Image *im);

                void setFirstFrame(int f);
                void setLastFrame(int f);

                // Set the current frame
                void setFrame(int f, sf::Sprite& sprite);

                // Frames advance right
                void setPosDirection();

                // Frames advance left
                void setNegDirection();
};

Animation::Animation() : currentFrame(1), rows(1), columns(1), frameWidth(0), frameHeight(0), firstFrame(1), lastFrame(1), direction(1), bounce(false)
{
}

void Animation::animate(sf::Sprite& sprite)
{
        // If it's been long enough since the frame was last changed
        if (delayTimer.GetElapsedTime() * 1000 > frameDelay)
        {
                // Increment frame left or right depending on the direction
                currentFrame += direction;

                // If we've exceed the last frame,
                if (currentFrame > lastFrame)
                {
                        if (bounce)
                        {
                                // Start moving left
                                direction = -1;

                                // Already made one step forward, so make two steps back
                                currentFrame -= 2;
                        }
                        else
                        {
                                // Cycle to the first frame
                                currentFrame = firstFrame;
                        }
                }
                // If we've gone below the limit but we weren't bouncing (just animating left)
                // loop back to the last frame.
                else if (currentFrame <= 0 && !bounce)
                {
                        currentFrame = lastFrame;
                }
                // If we've gone left of the first frame but want to bounce back
                else if (currentFrame <= 0 && bounce)
                {
                        // Reverse to positive direction
                        direction = 1;

                        // Already took one step back, so take two steps forward
                        currentFrame += 2;
                }
                else
                {
                        // Should never get here...
                }

                // Go to the new frame
                setFrame(currentFrame, sprite);

                delayTimer.Reset();
        }
}

void Animation::setRows(int r, const sf::Image *im)
{
        // Can't have 0 rows, or we'd end up dividing by 0 to get frameHeight
        if (r > 0)
        {
                rows = r;
        }

        frameHeight = im->GetHeight() / rows;
}

void Animation::setColumns(int c, const sf::Image *im)
{
        // Can't have 0 columns, or we'd end up dividing by 0 to get frameWidth
        if (c > 0)
        {
                columns = c;
        }

        frameWidth = im->GetWidth() / columns;
}

void Animation::setFirstFrame(int f)
{
        firstFrame = f;
}

void Animation::setLastFrame(int f)
{
        lastFrame = f;
}

void Animation::setFrame(int f, sf::Sprite& sprite)
{
        // Frame indices must start at 1
        if (f > 0)
        {
                currentFrame = f;

                // Current subrect
                sf::IntRect subRect = sprite.GetSubRect();

                unsigned int left = subRect.Left;
                unsigned int top = subRect.Top;
                unsigned int right = subRect.Right;
                unsigned int bottom = subRect.Bottom;

                int newRow, newCol;

                // Loop around column-wise
                if (currentFrame % columns == 0)
                {
                        newCol = columns;
                }
                else
                {
                        newCol = currentFrame - (columns * (currentFrame / columns));
                }

                // Find the appropriate row
                if (currentFrame % columns == 0)
                {
                        newRow = currentFrame / columns;
                }
                else
                {
                        newRow = (currentFrame / columns) + 1;
                }

                right = newCol * frameWidth;
                left = right - frameWidth;
                bottom = newRow * frameHeight;
                top = bottom - frameHeight;

                // Apply the changes
                sprite.SetSubRect(sf::IntRect(left, top, right, bottom));
        }
}

inline void Animation::setPosDirection()
{
        direction = 1;
}

inline void Animation::setNegDirection()
{
        direction = -1;
}

// The most generic sprite class with a few useful wrapper functions
class Sprite : public sf::Sprite
{
        protected:
                // Coordinates before the last movement
                float prevX;

                float prevY;

                sf::Rect<float> collisionRect;

        public:
                // Attributes

                // Stores attributes and methods for animating this sprite
                Animation animation;

                // Multipurpose identification code of the object
                int id;

                // x- and y-axis movement speeds
                float xMov;

                float yMov;

                // Methods

                Sprite();

                float getX();

                float getY();

                float getCenterX();

                float getCenterY();

                float getW();

                float getH();

                float getPrevX();

                float getPrevY();

                sf::Rect<float> &getCollisionRect();

                // Set the rect used for collision detection purposes only.
                // Note: this is not actually a rectangle, but rather defines the amount to add/subtract
                // from each side of the sprite's actual rect. 0 means no change.
                void setCollisionRect(float left, float top, float right, float bottom);

                // Move the sprite and save previous coordinates
                // (Replaces the use of sf::Sprite::Move)
                void moveAndSave(float deltaTime);
};

Sprite::Sprite() : collisionRect(0.f, 0.f, 0.f, 0.f), id(-1), xMov(0.f), yMov(0.f)
{
        sf::Vector2f position = GetPosition();
        prevX = position.x;
        prevY = position.y;
}

float Sprite::getX()
{
        return GetPosition().x;
}

float Sprite::getY()
{
        return GetPosition().y;
}

float Sprite::getCenterX()
{
        return GetPosition().x + (GetSize().x / 2);
}

float Sprite::getCenterY()
{
        return GetPosition().y + (GetSize().y / 2);
}

float Sprite::getW()
{
        return GetSize().x;
}

float Sprite::getH()
{
        return GetSize().y;
}

float Sprite::getPrevX()
{
        return prevX;
}

float Sprite::getPrevY()
{
        return prevY;
}

sf::Rect<float> &Sprite::getCollisionRect()
{
        return collisionRect;
}

void Sprite::setCollisionRect(float left, float top, float right, float bottom)
{
        collisionRect.Left = left;
        collisionRect.Top = top;
        collisionRect.Right = right;
        collisionRect.Bottom = bottom;
}

void Sprite::moveAndSave(float deltaTime)
{
        sf::Vector2f position = GetPosition();
        prevX = position.x;
        prevY = position.y;

        Move(xMov * deltaTime, yMov * deltaTime);
}

class Engine
{
        private:
                int screenWidth;

                int screenHeight;

                int bpp;

        protected:
                sf::RenderWindow window;

                // Log for writing error messages to.
                // This does not open until an error is written, but it stays open
                // until the game is over once it's opened.
                std::ofstream errorLog;

                // Are we in debug mode?
                bool debugOn;

                // Time to quit the level yet?
                bool quitLevel;

                // Have we initialized the current level yet?
                bool initializedLevel;

                // My IP addresses
                sf::IPAddress myGlobalIP;
                sf::IPAddress myLocalIP;

                // String input based on the most recent keypresses.
                // Usually grabbed for input and then flushed each game loop.
                std::string inputBuffer;

                // Used to display framerate
                sf::String framerateString;

        public:
                Engine();

                Engine(int screenWidth, int screenHeight, int bpp, unsigned long mode, std::string title);

                void setVideoMode(int screenWidth, int screenHeight, int bpp, unsigned long mode, std::string title);

                // Based on an event key code, add the appropriate character to the inputBuffer
                void addToInputBuffer(int keyCode);

                void writeError(std::stringstream& ss);

                void writeError(std::string s);

                float getScreenWidth();

                float getScreenHeight();

                void fillScreen(int r, int g, int b, int a);

                void printText(float x, float y, std::string text);

                // Print the framerate in the top right corner of the screen
                void printFramerate();

                ~Engine();
};

Engine::Engine(int _screenWidth, int _screenHeight, int _bpp, unsigned long mode, std::string title) : screenWidth(_screenWidth), screenHeight(_screenHeight), bpp(_bpp),
        debugOn(false), quitLevel(false), initializedLevel(false), framerateString("", sf::Font::GetDefaultFont(), 30)
{
        setVideoMode(screenWidth, screenHeight, bpp, mode, title);

        // Get my WAN and LAN IP addresses
        myGlobalIP = sf::IPAddress::GetPublicAddress();
        myLocalIP = sf::IPAddress::GetLocalAddress();
};

void Engine::setVideoMode(int screenWidth, int screenHeight, int bpp, unsigned long mode, std::string title)
{
        // Set the video mode
        sf::VideoMode vmode = sf::VideoMode(screenWidth, screenHeight, bpp);

        // Make sure the video mode is valid
        if (!vmode.IsValid())
        {
                std::stringstream ss;
                ss << "Error in Engine::setVideoMode: video mode not supported";
                writeError(ss);

                // Proceed with default video mode?
        }

        // Create the window with the desired settings
        window.Create(vmode, title, mode);

        // Window settings
        window.ShowMouseCursor(false);
}

void Engine::addToInputBuffer(int keyCode)
{
        if (keyCode == sf::Key::Num0)
        {
                inputBuffer += "0";
        }
        else if (keyCode == sf::Key::Num1)
        {
                inputBuffer += "1";
        }
        else if (keyCode == sf::Key::Num2)
        {
                inputBuffer += "2";
        }
        else if (keyCode == sf::Key::Num3)
        {
                inputBuffer += "3";
        }
        else if (keyCode == sf::Key::Num4)
        {
                inputBuffer += "4";
        }
        else if (keyCode == sf::Key::Num5)
        {
                inputBuffer += "5";
        }
        else if (keyCode == sf::Key::Num6)
        {
                inputBuffer += "6";
        }
        else if (keyCode == sf::Key::Num7)
        {
                inputBuffer += "7";
        }
        else if (keyCode == sf::Key::Num8)
        {
                inputBuffer += "8";
        }
        else if (keyCode == sf::Key::Num9)
        {
                inputBuffer += "9";
        }
        else if (keyCode == sf::Key::Period)
        {
                inputBuffer += ".";
        }
        // Special codes: if the input buffer is set to one of these,
        // an input handler can optionally perform a special action.

        // Backspace sets the input buffer to "backspace" so we know to remove a character
        else if (keyCode == sf::Key::Back)
        {
                inputBuffer = "backspace";
        }
        // Return notifies us to end the input
        else if (keyCode == sf::Key::Return)
        {
                inputBuffer = "return";
        }
        else if (keyCode == sf::Key::Left)
        {
                inputBuffer = "left arrow";
        }
        else if (keyCode == sf::Key::Right)
        {
                inputBuffer = "right arrow";
        }
}

void Engine::writeError(std::stringstream& ss)
{
        // Open the error log if it's not already open
        if (!errorLog.is_open())
        {
                errorLog.open("errors.txt");
        }

        // If the error log opened OK, write the error report
        if (!errorLog.fail())
        {
                std::string message = ss.str();
                message += "\n";

                errorLog << message;
        }
}

void Engine::writeError(std::string s)
{
        // Open the error log if it's not already open
        if (!errorLog.is_open())
        {
                errorLog.open("errors.txt");
        }

        // If the error log opened OK, write the error report
        if (!errorLog.fail())
        {
                s += "\n";

                errorLog << s;
        }
}

float Engine::getScreenWidth()
{
        return float(screenWidth);
}

float Engine::getScreenHeight()
{
        return float(screenHeight);
}

void Engine::fillScreen(int r, int g, int b, int a)
{
        window.Clear(sf::Color(r, g, b, a));
}

// NOTE: this function isn't incredibly efficient since it re-creates the sf::String
// variable every time it prints it.  It's generally better to create a static variable
// and print that, but this is more convenient when you don't want to do that.
void Engine::printText(float x, float y, std::string message)
{
        // Make the message the default font (Arial), and make the text size 30
        // to match the default character size
        sf::String text(message, sf::Font::GetDefaultFont(), 30);
        text.SetPosition(x, y);

        window.Draw(text);
}

void Engine::printFramerate()
{
        float framerate = 1.f / window.GetFrameTime();

        std::stringstream ss;
        ss << framerate;

        framerateString.SetText(ss.str());
        framerateString.SetPosition(screenWidth - framerateString.GetRect().GetWidth(), 0.f);

        window.Draw(framerateString);
}

Engine::~Engine()
{
        if (window.IsOpened())
                window.Close();

        if (errorLog.is_open())
        {
                errorLog.close();
        }
}

class KeyInput
{
        private:
                sf::Clock blinkTimer;

                int blinkDelay;

                sf::Clock inputTimer;

                int inputDelay;
                
                // Is the cursor filled?
                bool fillEnabled;

                // Message that has been input
                sf::String message;

                float height;

                float width;

                unsigned int characterLimit;

                sf::Shape cursor;

        public:

                KeyInput(float _width, float _height, int _blinkDelay, unsigned int _characterLimit);

                void blink();

                // Grabs keyboard input.
                // Returns true if the user signaled input completion
                bool getInput(std::string inputBuffer);

                void setCursorPosition(float x, float y);

                // Observers

                sf::Shape& getCursor();

                sf::String& getMessage();
};

KeyInput::KeyInput(float _width, float _height, int _blinkDelay, unsigned int _characterLimit) : fillEnabled(true), characterLimit(_characterLimit)
{
        width = _width;
        height = _height;

        cursor = sf::Shape::Rectangle(0.f, 0.f, width, height, sf::Color::White, 2.f, sf::Color::White);

        // Set how long it takes to blink the cursor
        blinkDelay = _blinkDelay;

        inputDelay = 80;
}

void KeyInput::blink()
{
        if (blinkTimer.GetElapsedTime() * 1000 > blinkDelay)
        {
                // Toggle the fill of the cursor shape
                cursor.EnableFill(!fillEnabled);
                fillEnabled = !fillEnabled;

                blinkTimer.Reset();
        }
}

bool KeyInput::getInput(std::string inputBuffer)
{
        // Current message
        std::string s = message.GetText();

        // Done inputting?
        bool done = false;

        // Special codes
        // If we're supposed to backspace, chop off a character
        if (inputBuffer == "backspace")
        {
                // Don't backspace unless there's at least one character
                if (s.length() > 0)
                {
                        s = s.substr(0, s.length() - 1);
                }
        }
        else if (inputBuffer == "return")
        {
                done = true;
        }
        // Normal input
        else
        {
                // Add newly typed characters
                s += inputBuffer;

                // Chop off characters if we exceeded the character limit
                if (s.length() > characterLimit)
                {
                        s = s.substr(0, characterLimit);
                }
        }

        // Set the updated text
        message.SetText(s);

        if (s.length() > 0)
        {
                // Set the cursor directly after the message
                cursor.SetPosition(message.GetPosition().x + message.GetRect().GetWidth(), cursor.GetPosition().y);
        }

        return done;
}

void KeyInput::setCursorPosition(float x, float y)
{
        cursor.SetPosition(x, y);

        // Message should be vertically alined with the text
        message.SetPosition(x, y);
}

sf::Shape& KeyInput::getCursor()
{
        return cursor;
}

sf::String& KeyInput::getMessage()
{
        return message;
}

class Selector
{
        private:
                // What are we polling for?
                std::string question;

                // Possible choices the player has
                std::vector<std::string> options;

                // Currently selected option
                unsigned int currentOption;

                // Arrows wrapped around the option so the player can page through
                sf::Shape leftArrow;

                sf::Shape rightArrow;

                // Text that stores the prompt (for displaying)
                sf::String prompt;

                // Text that stores the currently selected option (for displaying)
                sf::String option;

                // Private function used by constructors to set up arrow polygons
                void setShapes();

                // Private function used to change the currently displayed option;
                // usually called when the user presses an arrow to change options
                void setOption();

        public:
                Selector();

                Selector(std::string _question, std::vector<std::string> _options);

                void setQuestion(std::string _question);

                void setOptions(std::vector<std::string> _options);

                bool poll(std::string inputBuffer);

                void setPromptPosition(float x, float y);

                void setOptionPosition(float x, float y);

                // Align the arrows to the left and right of the text
                void alignArrows();

                // Observers

                int getCurrentOption();

                sf::String& getPrompt();

                sf::String& getOption();

                sf::Shape& getLeftArrow();

                sf::Shape& getRightArrow();
};

Selector::Selector() : question("Is this OK?"), currentOption(0)
{
        // Default options, should be overridden
        options.push_back("OK");

        setShapes();
}

Selector::Selector(std::string _question, std::vector<std::string> _options) : question(_question), currentOption(0), prompt(_question)
{
        options = _options;

        // No options?  Then give some defaults
        if (options.empty())
        {
                options.push_back("OK");
        }
        else
        {
                // Start at the first option
                option.SetText(options[0]);
        }

        setShapes();
}

void Selector::setShapes()
{
        leftArrow.AddPoint(0.f, 15.f, sf::Color(255, 255, 255), sf::Color(255, 255, 255));
        leftArrow.AddPoint(20.f, 30.f, sf::Color(255, 255, 255), sf::Color(255, 255, 255));
        leftArrow.AddPoint(20.f, 0.f, sf::Color(255, 255, 255), sf::Color(255, 255, 255));

        rightArrow.AddPoint(0.f, 0.f, sf::Color(255, 255, 255), sf::Color(255, 255, 255));
        rightArrow.AddPoint(0.f, 30.f, sf::Color(255, 255, 255), sf::Color(255, 255, 255));
        rightArrow.AddPoint(20.f, 15.f, sf::Color(255, 255, 255), sf::Color(255, 255, 255));
}

void Selector::setOption()
{
        option.SetText(options[currentOption]);

        // Size of the option text may be different, so we have to realign the arrows accordingly
        alignArrows();
}

void Selector::setQuestion(std::string _question)
{
        question = _question;

        prompt.SetText(question);
}

void Selector::setOptions(std::vector<std::string> _options)
{
        options = _options;

        if (!options.empty())
        {
                option.SetText(options[0]);
        }
}

bool Selector::poll(std::string inputBuffer)
{
        bool done = false;

        if (inputBuffer == "return")
        {
                done = true;
        }
        else if (inputBuffer == "left arrow")
        {
                if (currentOption > 0)
                {
                        currentOption--;

                        setOption();
                }
        }
        else if (inputBuffer == "right arrow")
        {
                if (currentOption < options.size() - 1)
                {
                        currentOption++;

                        setOption();
                }
        }

        return done;
}

void Selector::setPromptPosition(float x, float y)
{
        prompt.SetPosition(x, y);
}

void Selector::setOptionPosition(float x, float y)
{
        option.SetPosition(x, y);

        // Option text may have been moved, so move the arrows to line up
        alignArrows();
}

void Selector::alignArrows()
{
        // TODO: make this editable in the future?
        const float OFFSET = 10.f;

        float leftArrowWidth = leftArrow.GetPointPosition(1).x - leftArrow.GetPointPosition(0).x;

        // Align the shapes to the left and right of the text
        leftArrow.SetPosition(option.GetPosition().x - leftArrowWidth - OFFSET, option.GetPosition().y);
        rightArrow.SetPosition(option.GetPosition().x + option.GetRect().GetWidth() + OFFSET, option.GetPosition().y);
}

int Selector::getCurrentOption()
{
        return currentOption;
}

sf::String& Selector::getPrompt()
{
        return prompt;
}

sf::String& Selector::getOption()
{
        return option;
}

sf::Shape& Selector::getLeftArrow()
{
        return leftArrow;
}

sf::Shape& Selector::getRightArrow()
{
        return rightArrow;
}

// A sprite composed of many images.
// Usually used for backgrounds.
class BigSprite
{
        public:
                // An "object" is the pair composed of a sprite and the image associated with it
                std::vector< std::pair<sf::Image*, Sprite*> > objects;

                BigSprite();

                ~BigSprite();

                // Append a sprite to the right side of the big sprite.
                // Returns true if the image loaded properly, false otherwise.
                bool addSprite(std::string filename);

                // Observers

                Sprite *getSprite(unsigned int index);

                // Total width
                float getW();
};

BigSprite::BigSprite()
{
}

bool BigSprite::addSprite(std::string filename)
{
        sf::Image *im = new sf::Image;

        Sprite *sp = new Sprite;

        if (!im->LoadFromFile(filename))
        {
                // Couldn't load the image.  Return false for error processing purposes
                return false;
        }

        im->SetSmooth(false);

        sp->SetImage(*im);

        if (!objects.empty())
        {
                // Object directly to the left of this new one
                Sprite *leftSprite = objects.back().second;
                sp->SetPosition(leftSprite->GetPosition().x + leftSprite->GetSize().x, leftSprite->GetPosition().y);
        }

        std::pair<sf::Image*, Sprite*> obj(im, sp);

        objects.push_back(obj);

        return true;
}

Sprite *BigSprite::getSprite(unsigned int index)
{
        if (index >= 0 && index < objects.size())
        {
                return objects[index].second;
        }
        else
        {
                // Bad index!
                return NULL;
        }
}

float BigSprite::getW()
{
        if (!objects.empty())
        {
                Sprite *leftSprite = objects[0].second;
                Sprite *rightSprite = objects.back().second;
                return (rightSprite->GetPosition().x + rightSprite->GetSize().x) - leftSprite->GetPosition().x;
        }
        else
        {
                return 0.f;
        }
}

BigSprite::~BigSprite()
{
        // Free up memory allocated for images
        for (unsigned int i = 0; i < objects.size(); i++)
        {
                // FIXME: triggered a breakpoint.  Problem here?
                delete objects[i].first;
                delete objects[i].second;
        }
}
