#include "engine.h"
#include <math.h>
#include <queue>
#include <map>

// Used to index adjacent rooms
enum {LEFT, RIGHT, UP, DOWN};

// Player states
enum {FACING_RIGHT, FACING_LEFT};

// Monster states
enum {WANDERING, HUNTING, ATTACKING, DEAD};

// Used for conversion
const float RADIANS_TO_DEGREES = 57.2957f;

// Minimum of two ints
int min(int a, int b);

// Return true if the element is in the vector
bool inVector(int i, std::vector<int>& v);

class Room
{
        private:
        public:
                // What files to stitch the image together from
                std::vector<std::string> filenames;

                // Background image
                BigSprite bg;                

                // Links to other rooms
                Room *links[4];

                bool hasLeftDoor, hasRightDoor;
                bool hasLeftWall, hasRightWall;

                // Boundary lines in the room the player can't cross
                // NOTE: rightBoundary should be subtracted manually from the background width.
                float leftBoundary, rightBoundary;

                // Stairwell up left and right boundary
                sf::Rect<float> stairsUp;

                // Stairwell down left and right boundary
                sf::Rect<float> stairsDown;

                // Contents

                Room();

                // Automatically load the background tiles based on the filenames.
                // Returns false if there was an error loading
                bool loadBackground();

                // Returns true if the background has at least one tile loaded.
                // This is to help us avoid loading background multiple times.
                bool hasBackground();
};

Room::Room() : hasLeftDoor(false), hasRightDoor(false), hasLeftWall(false), hasRightWall(false), leftBoundary(0.f), rightBoundary(0.f)
{
        // No links until we initialize the map
        links[0] = NULL;
        links[1] = NULL;
        links[2] = NULL;
        links[3] = NULL;
}

bool Room::loadBackground()
{
        bool finishedOkay = true;

        // For each filename, load that image into the background tiles
        for (unsigned int i = 0; i < filenames.size() && finishedOkay; i++)
        {
                // The full path (including resource dir)
                std::string path = "./rc/game/";
                path += filenames[i];

                if (!bg.addSprite(path))
                {
                        finishedOkay = false;
                }
        }

        return finishedOkay;
}

inline bool Room::hasBackground()
{
        return !bg.objects.empty();
}

class Pickup : public Sprite
{
        public:
                Room *parentRoom;

                static sf::Image *image;
};
sf::Image *Pickup::image = NULL;

// Playable character
class Player : public Sprite
{
        protected:
                // Current state of action, e.g. running or jumping
                int state;

                // True if the state was changed to a new one,
                // false if the last state change did nothing
                bool changedState;

        public:
                // How much life the player has left
                int health;

                // Amount of ammo that can be shot before reloading
                int clipSize;

                // Amount of ammunition left
                int ammo;

                // Amount of ammunition in the gun
                int clip;

                // Used to display endgame stats
                int shotsHit;
                int shotsMissed;
                int monstersKilled;

                // Room the player is staying in
                Room *parentRoom;

                void setState(int newState);

                int getState();

                bool getChangedState();
};

void Player::setState(int newState)
{
        if (state != newState)
        {
                changedState = true;
        }
        else
        {
                changedState = false;
        }

        state = newState;
}

int Player::getState()
{
        return state;
}

bool Player::getChangedState()
{
        return changedState;
}

class Portal : public Sprite
{
        private:
                static const float ROTATION_AMOUNT;

        public:
                Portal();

                int maxHealth;
                int health;

                int lives;
                
                Room *parentRoom;

                // Controls the rate of monster spawning
                sf::Clock spawnTimer;
                int spawnDelay;

                // Spinning effect
                void rotate(float deltaTime);
};

Portal::Portal() : maxHealth(10), spawnDelay(0), lives(3)
{
        health = maxHealth;
}

void Portal::rotate(float deltaTime)
{
        Rotate(ROTATION_AMOUNT * deltaTime);
}

// A monster in the game
// Not actually playable, but inherits state functions from Player
// that allow it to have different behavior cycles
class Monster : public Player
{
        private:
                static const float MIN_SPEED;
                static const float MAX_SPEED;

                // Some rare monsters will move much faster than others
                static const float SUPER_SPEED;

                // Monsters have a 1 in CHANCE_OF_SUPER possibility of becoming a super-monster
                static const int CHANCE_OF_SUPER = 10;

                static const float ANIMATION_DELAY;

                // How much life the monster has left
                int health;

        public:
                int maxHealth;

                // Controls how frequently the monster can attack
                sf::Clock attackTimer;
                int attackDelay;

                Monster();

                // Room we are trying to move to in HUNTING stage.
                // Once in this room, we can begin to move toward the closest player.
                Room *targetRoom;

                // Image for every monster
                static sf::Image *image;

                // Move the monster and reflect it if it changes direction
                void move(float deltaTime);

                void hurt(int damage);

                int getHealth();
};
sf::Image *Monster::image = NULL;

Monster::Monster() : maxHealth(5), attackDelay(700)
{
        health = maxHealth;

        // Move at slightly randomized speed
        xMov = sf::Randomizer::Random(MIN_SPEED, MAX_SPEED);

        // 1 in 10 chance of becoming a super monster!
        if (sf::Randomizer::Random(1, CHANCE_OF_SUPER) == CHANCE_OF_SUPER)
        {
                xMov = SUPER_SPEED;
        }

        animation.frameDelay = ANIMATION_DELAY;
        animation.bounce = true;

        // We're not in a room or hunting for a room yet
        parentRoom = NULL;
        targetRoom = NULL;

        // Assume we haven't found a player to hunt yet
        state = WANDERING;
}

void Monster::move(float deltaTime)
{
        // Flip the character if they're moving left (the sprite is by default oriented to the right
        if (xMov < 0)
        {
                FlipX(true);
        }
        else if (xMov > 0)
        {
                FlipX(false);
        }

        Move(xMov * deltaTime, yMov * deltaTime);
}

void Monster::hurt(int damage)
{
        // Take damage
        health -= damage;

        if (health <= 0)
        {
                health = 0;
                setState(DEAD);
        }
}

inline int Monster::getHealth()
{
        return health;
}

class Game : public Engine
{
        private:
                // Vector of functions that define how to initialize, update, and display the game.
                typedef void (Game::*state)();
                std::queue<state> gameStates;

                // Are we the host or client?
                enum {HOST, CLIENT, DISCONNECTED};
                int networkStatus;

                // What network mode are we using?
                enum {SINGLEPLAYER, MULTIPLAYER};
                int networkMode;

                // Second player's IP address
                sf::IPAddress theirIP;

                // The main camera
                sf::View camera;

                // Have we played the game once already?
                bool playedBefore;

                // Number of pickups to spawn
                static const int MAX_PICKUPS = 5;

                // Amount of ammo increased by pickup
                static const int AMMO_PICKUP_INCREASE = 48;

                // Width and height of frames for both players
                static const int PLAYER_FRAME_WIDTH = 185;
                static const int PLAYER_FRAME_HEIGHT = 300;

                // Rows/columns in player animations
                static const int PLAYER_ROWS = 5;
                static const int PLAYER_COLUMNS = 2;

                // Frame boundaries in player animation
                static const int PLAYER_LAST_FRAME = 9;
                static const int PLAYER_FIRST_FRAME = 2;

                // Animation delay of the players (the slower this is, the faster they'll appear to run)
                static const float PLAYER_ANIMATION_DELAY;
                static const float PLAYER_ANIMATION_DELAY_SLOW;

                // How much damage the players do
                static const int PLAYER_DAMAGE = 1;
                // How much they can take
                static const int PLAYER_HEALTH = 10;
                // How much ammo to start with
                static const int PLAYER_AMMO = 48;

                // Running speed of the player        
                static const float RUN_SPEED;

                // Speed when walking backward
                static const float BACKWARD_SPEED;

                static const int MONSTER_FRAME_WIDTH = 144;
                static const int MONSTER_FRAME_HEIGHT = 288;
                static const int MONSTER_ROWS = 2;
                static const int MONSTER_COLUMNS = 3;
                static const int MONSTER_FIRST_FRAME = 2;
                static const int MONSTER_LAST_FRAME = 6;
                static const float MONSTER_ANIMATION_DELAY;

                // Y-axis position of players
                static const float FLOOR_HEIGHT;

                // Rotation center for front and back arms
                static const float ARM_ROTATION_CENTER_X;
                static const float ARM_ROTATION_CENTER_Y;

                // Max and min arm angles
                static const float MAX_AIM_ANGLE;
                static const float MIN_AIM_ANGLE;

                // Amount to rotate the arm by due to recoil
                static const float RECOIL_ANGLE;

                int livingMonsters;

                // Did the player win the game?
                bool gameWon;

                // Player 1's image and sprite
                sf::Image _myPlayer;
                Player myPlayer;

                // Player 1's back (obscured) and front arms
                sf::Image _myPlayerBArm;
                Sprite myPlayerBArm;
                sf::Image _myPlayerFArm;
                Sprite myPlayerFArm;

                // Player 2's image and sprite
                sf::Image _theirPlayer;
                Player theirPlayer;

                // Portal sprite
                Portal portal;
                sf::Image _portal;

                // All rooms in the game
                std::map<std::string, Room*> rooms;

                // List of all the monsters spawned in the game
                std::vector<Monster*> monsters;

                // List of pickups in the game
                std::vector<Pickup*> pickups;

                // Used to spawn the initial monsters randomly throughout the house, and from there we only
                // spawn them from portals.
                bool initializedMonsters;

        public:
                Game(int screenWidth, int screenHeight, int bpp, int mode, std::string title);

                ~Game();

                // Handle the current process and manage the queue of processes
                void gameMain();

                // LEVELS

                // Let the player choose the game type (single player, LAN, WAN)
                void gameSelection();

                void gameInstructions();

                void intro();
                
                // Lobby where the host waits for a client to connect
                void hostLobby();

                // Lobby where the client can connect to a host
                void clientLobby();

                // Main game: for playing multiplayer as host or playing singleplayer
                void hostGame();

                // Main game: for the client player
                void clientGame();

                // Game over screen
                void gameOver();

                // Cutscene that plays when the last portal is killed
                void finalCutscene();

                // Helper functions

                // Initialize adjacency lists of rooms
                // to the room the player will start in
                void initializeMap();

                // Delete all pointers that comprise the map
                void deleteMap();

                // Distance between two points
                float distance(float x1, float y1, float x2, float y2);

                void handleCameras(float levelWidth);

                void handlePlayerControls(int directionFacing);

                void resetCamera();

                void spawnMonsters(bool initializedMonsters);

                void deleteMonsters();

                void deletePickups();

                // Simply moves the monsters with no regard to their behavior
                // or current state
                void moveAndAnimateMonsters(float deltaTime);

                // Keeps monsters from going through walls, and allows them to
                // pass through doors.  Also handles AI.
                void monsterAI();
};

Game::Game(int screenWidth, int screenHeight, int bpp, int mode, std::string title) : Engine(screenWidth, screenHeight, bpp, mode, title)
{
        // Window settings

        window.SetFramerateLimit(60);

        // Set the queue of game states to step through
        gameStates.push(&Game::gameSelection);
        gameStates.push(&Game::gameInstructions);
        gameStates.push(&Game::intro);

        // Start out in single player until the player chooses to connect
        networkStatus = DISCONNECTED;
        networkMode = SINGLEPLAYER;

        // Seed the random number generate
        sf::Randomizer::SetSeed(time(NULL));

        // Set main camera and its parameters
        resetCamera();

        // Set up the map
        initializeMap();

        // Load portal background
        if (!_portal.LoadFromFile("./rc/game/portal.png"))
        {
                writeError("Couldn't load portal.png");
        }
        _portal.SetSmooth(false);
        portal.SetImage(_portal);

        // Pick a random room to put the portal in,
        // repeat until we have a random room that no player is in
        std::map<std::string, Room*>::iterator it;
        do
        {
                it = rooms.begin();

                int randRoom = sf::Randomizer::Random(0, rooms.size() - 1);

                // Iterate through the map of rooms until we have the randomly selected room
                for (int i = 0; i < randRoom && it != rooms.end(); i++)
                { 
                        it++;
                }
        } while (it->second == myPlayer.parentRoom);

        portal.SetCenter(portal.getW() / 2.f, portal.getH() / 2.f);

        // Place the portal in the center of the room
        portal.SetPosition(it->second->bg.getW() / 2.f, FLOOR_HEIGHT - (portal.getH() / 2.f) + 80.f);

        portal.parentRoom = it->second;

        portal.spawnDelay = 15;

        livingMonsters = 0;

        gameWon = false;

        initializedMonsters = false;

        sf::Image *_pickup = new sf::Image;

        if (!_pickup->LoadFromFile("./rc/game/pickup.png"))
        {
                writeError("Couldn't load pickup.png");
                quitLevel = true;
        }
        _pickup->SetSmooth(false);
        Pickup::image = _pickup;

        // Spawn the pickups, but don't spawn two in the same room
        for (unsigned int i = 0; i < MAX_PICKUPS; i++)
        {
                Pickup *p = new Pickup;

                // Pick a random room to put the pickup in,
                // repeat until we have a random room that no player is in
                int randRoom;
                std::map<std::string, Room*>::iterator it;
                do
                {
                        it = rooms.begin();

                        randRoom = sf::Randomizer::Random(0, rooms.size() - 1);

                        // Iterate through the map of rooms until we have the randomly selected room
                        for (int i = 0; i < randRoom && it != rooms.end(); i++)
                        { 
                                it++;
                        }
                } while (it->second == myPlayer.parentRoom || it->first == "Stairwell 1f" || it->first == "Stairwell 2f" || it->first == "Stairwell 3f");

                p->parentRoom = it->second;

                p->SetImage(*Pickup::image);

                p->SetPosition(200.f, FLOOR_HEIGHT - p->getH());

                pickups.push_back(p);
        }
        
        // This is our first time playing
        playedBefore = false;
}

Game::~Game()
{
        // Free up memory reserved for images and sprites

        deleteMap();

        deleteMonsters();

//        deletePickups();
}

void Game::gameMain()
{
        // Game loop
        while (!quitLevel && window.IsOpened() && !gameStates.empty())
        {
                // Process events
                sf::Event windowEvent;
                while (window.GetEvent(windowEvent))
                {
                        // Close window : exit
                        if (windowEvent.Type == sf::Event::Closed)
                        {
                                window.Close();
                        }

                        // A key has been pressed
                        if (windowEvent.Type == sf::Event::KeyPressed)
                        {
                                // Escape key: exit
                                if (windowEvent.Key.Code == sf::Key::Escape)
                                {
                                        window.Close();
                                }

                                // F1 key: take a screenshot
                                if (windowEvent.Key.Code == sf::Key::F1)
                                {
                                        sf::Image sshot = window.Capture();
                                        sshot.SaveToFile("screenshot.png");
                                }

                                // Tilde key: turn debug mode on/off
                                if (windowEvent.Key.Code == sf::Key::Tilde)
                                {
                                        if (!debugOn) debugOn = true;
                                        else debugOn = false;
                                }

                                addToInputBuffer(windowEvent.Key.Code);
                        }
                }

                // Execute the foremost game state
                (this->*gameStates.front())();

                // Apply all drawing updates
                window.Display();

                // Flush out the input buffer
                inputBuffer = "";
        }

        if (!gameStates.empty())
        {
                // Move on to the next state
                gameStates.pop();

                // If there are more states to execute, keep going!
                if (!gameStates.empty())
                {
                        // Make sure we keep playing
                        quitLevel = false;

                        // Will have to load resources for the next level
                        initializedLevel = false;

                        gameMain();
                }
        }
}

void Game::gameSelection()
{
        // Background
        static Sprite bg;
        static sf::Image _bg;

//        static Selector choice;

        static const float TEXT_INDENT = 20.f;

        static sf::Clock stagger;
        static const int STAGGER = 1;

        if (!initializedLevel)
        {
                if (!_bg.LoadFromFile("./rc/intro/titlescreen.png"))
                {
                        writeError("Couldn't load titlescreen.png");
                        quitLevel = true;
                }
                _bg.SetSmooth(false);
                bg.SetImage(_bg);

                stagger.Reset();

//                choice.setQuestion("Select a gametype");

//                std::vector<std::string> options;
//                options.push_back("Singleplayer");
//                options.push_back("Multiplayer - Host");
//                options.push_back("Multiplayer - Client");

//                choice.setOptions(options);

//                // Align prompt and option display text
//                float promptWidth = choice.getPrompt().GetRect().GetWidth();
//                float promptHeight = choice.getPrompt().GetRect().GetHeight();
//                choice.setPromptPosition((getScreenWidth() / 2.f) - (promptWidth / 2.f), 200.f);
//                float promptX = choice.getPrompt().GetPosition().x;
//                float promptY = choice.getPrompt().GetPosition().y;
//                choice.setOptionPosition(promptX + 40.f, promptY + 40.f);

                initializedLevel = true;
        }

        if (!playedBefore || (playedBefore && stagger.GetElapsedTime() > STAGGER))
        {
                if (window.GetInput().IsKeyDown(sf::Key::Return))
                {
                        networkMode = SINGLEPLAYER;
                        quitLevel = true;
                        gameStates.push(&Game::hostGame);
                }
        }

        // Allow the user to select an input, and see if they pressed enter
//        if (choice.poll(inputBuffer))
//        {
//                // What the player chose
//                std::string finalOption = choice.getOption().GetText();

//                // TODO: make a logical choice based on player input
//                if (finalOption == "Singleplayer")
//                {
//                        networkMode = SINGLEPLAYER;

//                        // Enter singleplayer mode
//                        gameStates.push(&Game::hostGame);

//                        quitLevel = true;
//                }
//                else if (finalOption == "Multiplayer - Host")
//                {
//                        networkMode = MULTIPLAYER;

//                        // Enter the host lobby, then show the intro
//                        gameStates.push(&Game::hostLobby);
////                        gameStates.push(&Game::intro);
//                        gameStates.push(&Game::hostGame);

//                        quitLevel = true;
//                }
//                else if (finalOption == "Multiplayer - Client")
//                {
//                        networkMode = MULTIPLAYER;

//                        // Enter the client lobby, then show the intro
//                        gameStates.push(&Game::clientLobby);
////                        gameStates.push(&Game::intro);
//                        gameStates.push(&Game::clientGame);

//                        quitLevel = true;
//                }
//                else
//                {
//                        // Invalid input
//                        writeError("Error in gameSelection: invalid input");
//                        quitLevel = true;
//                }
//        }

        window.Draw(bg);

//        window.Draw(choice.getPrompt());
//        window.Draw(choice.getOption());
//        window.Draw(choice.getLeftArrow());
//        window.Draw(choice.getRightArrow());

        if (quitLevel && playedBefore)
        {
                // RESET everything

                // Player is facing right by default
                myPlayer.setState(FACING_RIGHT);

                myPlayer.health = PLAYER_HEALTH;
                theirPlayer.health = PLAYER_HEALTH;

                myPlayer.clipSize = 12;
                theirPlayer.clipSize = 12;

                myPlayer.ammo = PLAYER_AMMO;
                myPlayer.clip = min(myPlayer.ammo, myPlayer.clipSize);
                theirPlayer.ammo = PLAYER_AMMO;

                myPlayer.shotsHit = 0;
                myPlayer.shotsMissed = 0;
                myPlayer.monstersKilled = 0;

                myPlayer.xMov = 0;

                myPlayer.parentRoom = rooms["Stairwell 1f"];

                // Set player 1's position
                myPlayer.SetPosition((getScreenWidth() / 2.f) - (myPlayer.getW() / 2.f), FLOOR_HEIGHT - myPlayer.getH());
                myPlayerBArm.SetPosition(myPlayer.getX() + ARM_ROTATION_CENTER_X, myPlayer.getY() + ARM_ROTATION_CENTER_Y);
                myPlayerFArm.SetPosition(myPlayer.getX() + ARM_ROTATION_CENTER_X, myPlayer.getY() + ARM_ROTATION_CENTER_Y);

                // Pick a random room to put the portal in,
                // repeat until we have a random room that no player is in
                std::map<std::string, Room*>::iterator it;
                do
                {
                        it = rooms.begin();

                        int randRoom = sf::Randomizer::Random(0, rooms.size() - 1);

                        // Iterate through the map of rooms until we have the randomly selected room
                        for (int i = 0; i < randRoom && it != rooms.end(); i++)
                        { 
                                it++;
                        }
                } while (it->second == myPlayer.parentRoom || it->first == "Stairwell 1f" || it->first == "Stairwell 2f" || it->first == "Stairwell 3f");

                // Place the portal in the center of the room
                portal.SetPosition(it->second->bg.getW() / 2.f, FLOOR_HEIGHT - (portal.getH() / 2.f) + 80.f);

                portal.parentRoom = it->second;

                portal.spawnTimer.Reset();

                portal.lives = 3;

                // Need to respawn monsters if we've played before
                deleteMonsters();

                livingMonsters = 0;

                // Clear out old pickups
                for (unsigned int i = 0; i < pickups.size(); i++)
                {
                        delete pickups[i];
                        pickups[i] = NULL;
                }

                for (unsigned int i = 0; i < pickups.size(); i++)
                {
                        Pickup *p = new Pickup;

                        // Pick a random room to put the pickup in,
                        // repeat until we have a random room that no player is in
                        int randRoom;
                        std::map<std::string, Room*>::iterator it;
                        do
                        {
                                it = rooms.begin();

                                randRoom = sf::Randomizer::Random(0, rooms.size() - 1);

                                // Iterate through the map of rooms until we have the randomly selected room
                                for (int i = 0; i < randRoom && it != rooms.end(); i++)
                                { 
                                        it++;
                                }
                        } while (it->second == myPlayer.parentRoom);

                        p->parentRoom = it->second;

                        p->SetImage(*Pickup::image);

                        p->SetPosition(200.f, FLOOR_HEIGHT - p->getH());

                        pickups[i] = p;
                }

                initializedMonsters = false;
        }
}

void Game::gameInstructions()
{
        static sf::Image _instructions;
        static Sprite instructions;

        static sf::Clock stagger;
        static const int STAGGER = 1;

        if (!initializedLevel)
        {
                if (!_instructions.LoadFromFile("./rc/intro/instructions.png"))
                {
                        writeError("Couldn't load instructions.png");
                        quitLevel = true;
                }
                _instructions.SetSmooth(false);
                instructions.SetImage(_instructions);

                initializedLevel = true;
        }

        if (stagger.GetElapsedTime() > STAGGER && window.GetInput().IsKeyDown(sf::Key::Return))
        {
                quitLevel = true;
        }

        window.Draw(instructions);
}

// Tries to open up a connection to a host socket
class TCPClientThread : public sf::Thread
{
        private:
                // Connection with a host
                sf::SocketTCP toHost;

                // Has Run() terminated yet?
                bool finished;

                // Has a connection been made?
                bool connected;

                static const unsigned short PORT = 4567;

                virtual void Run();

        public:
                TCPClientThread();

                virtual ~TCPClientThread();

                // Address to connect to
                sf::IPAddress hostIP;

                // Observers

                bool isFinished();

                bool isConnected();
};

TCPClientThread::TCPClientThread() : finished(true), connected(false)
{
}

TCPClientThread::~TCPClientThread()
{
        toHost.Close();
}

void TCPClientThread::Run()
{
        // Thread won't be finished until this thread terminates
        finished = false;

        if (toHost.Connect(PORT, hostIP) != sf::Socket::Done)
        {
                // Error connecting...
                connected = false;
        }
        else
        {
                // Successful connection!
                connected = true;
        }

        finished = true;
}

bool TCPClientThread::isFinished()
{
        return finished;
}

bool TCPClientThread::isConnected()
{
        return connected;
}

void Game::clientLobby()
{      
        // How much to indent text on this screen by
        const float TEXT_INDENT = 20.f;

        static sf::String prompt;

        static KeyInput input(20.f, 30.f, 600, 15);

        static sf::Image _bg;
        static Sprite bg;

        // Tries to set up a connection with the host
        static TCPClientThread client;

        // Messages to display to the player

        // Tell a client their connection with the host is pending
        static sf::String waitingHeader("Waiting for host...", sf::Font::GetDefaultFont(), 30);

        if (!initializedLevel)
        {
                prompt.SetText("Enter IP to connect (or wait for connection):"); 
                prompt.SetPosition(TEXT_INDENT, 30.f);

                input.setCursorPosition(TEXT_INDENT, prompt.GetPosition().y + prompt.GetRect().GetHeight() + 20.f);

                _bg.LoadFromFile("./rc/lobby/lobby-bg.png");
                _bg.SetSmooth(false);
                bg.SetImage(_bg);

                initializedLevel = true;
        }

        // If we're not currently trying to make a connection, get an IP from the client player
        if (client.isFinished())
        {
                input.blink();

                // If we're not already trying to connect, get input and check if user is finished
                if (input.getInput(inputBuffer))
                {
                        std::string s = input.getMessage().GetText();
                        theirIP = s;
                        client.hostIP = theirIP;

                        if (theirIP.IsValid())
                        {
                                // Try to connect
                                client.Launch();
                        }
                        else
                        {
                                // TODO: make this an actual onscreen message
                                writeError("Client entered invalid IP address");
                                quitLevel = true;
                        }
                }
        }

        // If we connected successfully to the host,
        if (client.isFinished() && client.isConnected())
        {
                // They'll grab our IP address for future use;
                // we can now start the game
                quitLevel = true;

                networkStatus = CLIENT;
        }

        window.Draw(bg);

        window.Draw(prompt);

        window.Draw(input.getMessage());

        window.Draw(input.getCursor());

        // If we're trying to connected but haven't successfully connected,
        // tell the client we're still waiting
        if (!client.isFinished() && !client.isConnected())
        {
                waitingHeader.SetPosition(TEXT_INDENT, input.getMessage().GetPosition().y + 50.f);

                window.Draw(waitingHeader);
        }

        if (debugOn)
        {
                printFramerate();

                std::stringstream ss;
                ss << client.isFinished() << " " << client.isConnected();
                printText(TEXT_INDENT, 400.f, ss.str());
        }

        // If a quit message was posted, kill the thread
        if (quitLevel)
        {
                client.Wait();
        }
}

// Thread for a TCP socket that listens for connections
class TCPListenerThread : public sf::Thread
{
        private:
                static const unsigned short PORT = 4567;

                // Is this thread finished running?
                bool finished;

                // Listens to a given port
                sf::SocketTCP listener;

                // Actual connection established through the port we're listening to
                sf::SocketTCP connection;

                // IP of the person we've connected to
                sf::IPAddress theirIP;

                virtual void Run();

        public:
                TCPListenerThread();

                virtual ~TCPListenerThread();

                // Observers

                sf::IPAddress getTheirIP();

                bool isFinished();
};

TCPListenerThread::TCPListenerThread() : finished(true)
{
}

TCPListenerThread::~TCPListenerThread()
{
        // Close the connection
        listener.Close();
        connection.Close();
}

void TCPListenerThread::Run()
{
        // Thread isn't finished until this function completes
        finished = false;

        // Start listening for connections
        if (!listener.Listen(PORT))
        {
                // Error...
        }

        // Wait for a connection
        if (listener.Accept(connection, &theirIP) != sf::Socket::Done)
        {
                // Error...
        }

        // Finished connecting
        finished = true;
}

sf::IPAddress TCPListenerThread::getTheirIP()
{
        return theirIP;
}

bool TCPListenerThread::isFinished()
{
        return finished;
}

void Game::hostLobby()
{      
        // How much to indent text on this screen by
        const float TEXT_INDENT = 20.f;

        static Selector connectionAccepter;

        static sf::Image _bg;
        static Sprite bg;

        // Will listen for a connection, then we'll fetch the other person's IP
        static TCPListenerThread host;

        // Messages to display to the player

        // Shows the player their global/local IP
        static sf::String playerIPHeader("", sf::Font::GetDefaultFont(), 30);

        if (!initializedLevel)
        {
                std::string s;
                s = "Your global IP: ";
                s += myGlobalIP.ToString();
                s += "\nYour local IP: ";
                s += myLocalIP.ToString();
                s += "\n\nAwaiting connection...";
                playerIPHeader.SetText(s);

                playerIPHeader.SetPosition(TEXT_INDENT, 20.f);

                _bg.LoadFromFile("./rc/lobby/lobby-bg.png");
                _bg.SetSmooth(false);
                bg.SetImage(_bg);

                // Listen for and accept a connection from another player
                // TODO: test for connection failure?
                host.Launch();

                // Settings for the connection acceptance prompt
                connectionAccepter.setQuestion("Accept connection from x?");
                std::vector<std::string> options;
                options.push_back("Yes");
                options.push_back("No");
                connectionAccepter.setOptions(options);

                // Position of the acceptance prompt
                connectionAccepter.setPromptPosition(TEXT_INDENT, playerIPHeader.GetPosition().y + playerIPHeader.GetRect().GetHeight());
                sf::String p = connectionAccepter.getPrompt();
                connectionAccepter.setOptionPosition(TEXT_INDENT + 30.f, p.GetPosition().y + p.GetRect().GetHeight());

                initializedLevel = true;
        }

        window.Draw(bg);

        if (host.isFinished())
        {
                // Allow the player to accept or deny connection
                // TODO: add this later.  Too complicated for now.

//                std::string question = "Accept connection from ";
//                question += host.getTheirIP().ToString();
//                connectionAccepter.setQuestion(question);
//                question += "?";

//                if (connectionAccepter.poll(inputBuffer))
//                {
//                        std::string choice = connectionAccepter.getOption().GetText();

//                        if (choice == "Yes")
//                        {
//                                // Store the IP of the person we connected to
//                                theirIP = host.getTheirIP().ToString();

//                                // Ready to move onto the main game
//                                quitLevel = true;
//                        }
//                        else if (choice == "No")
//                        {
//                                // Start listening for a connection again
//                                host.Launch();
//                        }
//                        else
//                        {
//                                writeError("Undefined choice in hostLobby");
//                                quitLevel = true;
//                        }
//                }
                theirIP = host.getTheirIP().ToString();

                quitLevel = true;

                networkStatus = HOST;
        }

//        if (host.isFinished())
//        {
//                window.Draw(connectionAccepter.getPrompt());
//                window.Draw(connectionAccepter.getOption());
//                window.Draw(connectionAccepter.getLeftArrow());
//                window.Draw(connectionAccepter.getRightArrow());
//        }

        // print my public and local IP
        window.Draw(playerIPHeader);

        if (debugOn)
        {
                printFramerate();

                if (!host.isFinished())
                {
                        printText(TEXT_INDENT, 400.f, "Thread is not finished");
                }
                else
                {
                        printText(TEXT_INDENT, 400.f, "Thread is finished");
                }
        }

        // If a quit message was posted, kill the thread
        if (quitLevel)
        {
                host.Wait();
        }
}

void Game::intro()
{
        // Words we will flash in the intro
        static std::vector<sf::String> words;

        // Used to measure the time until another word pops up
        static sf::Clock wordDisplayTimer;
        static int wordDisplayDelay = 500;

        // The max number of words we're displaying.
        // We steadily increase this number until all words are shown
        static unsigned int maxWordsShown = 0;

        // Music that plays after sound is done
        static sf::Music introMusic;

        if (!initializedLevel)
        {
                // Phase 1
                words.push_back(sf::String("alchemy", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("collapse", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("close", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("flesh", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("seems", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("singularity", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("melting", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("recursive", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("fractal", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("strange loop", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("too far", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("time", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("ripping", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("escape", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("invasion", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("experiment", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("eternal recurrence", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("putrid", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("dripping", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("eidolon", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("horror", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("and for what?", sf::Font::GetDefaultFont(), 30));

                // Phase 2
                words.push_back(sf::String("ALCHEMY", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("cOllAPse", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("clOse", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("FleSH", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("sEEms", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("sINgUlARIty", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("MeLTING", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("ReCURSIVE", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("FrAcTAL", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("STRA[]GE loop", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("too far!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("txxx", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("r!pp!ng", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("3scap3", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("inv_sion", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("EXperiment", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("exxxxxx recuXXence", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("pUtr!d", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("dr_pp_ng", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("e[]dolon", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("h0rr0r", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("and for what?", sf::Font::GetDefaultFont(), 30));

                // Phase 3
                words.push_back(sf::String("ALCHEMY", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("RIPPING, SHREDDING", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("TORN APART", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("foundation,disappearing", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("no_sense_anymore", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("?????", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("XXXXXX!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("recurecurecur", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("frXctXl", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("...s.r.nge l..p...", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("TOO FAR", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("!!!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("a sound, not far off", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("DIsturBANCE", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("conf?sion", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("ALCHEMY?", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("ALCHEMY!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("mad,", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("mad science", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("ABsURD!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("ripping,", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("...shreddING", sf::Font::GetDefaultFont(), 30));

                // Phase 4
                words.push_back(sf::String("FLESH SALAD", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("eventHORIZON", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("BLACK SWAN", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("events beyond our control", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("NO SENSE ANYMORE", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));
                words.push_back(sf::String("HELP!", sf::Font::GetDefaultFont(), 30));

                float screenWidth = getScreenWidth();
                float screenHeight = getScreenHeight();

                for (unsigned int i = 0; i < words.size(); i++)
                {
                        // Set a random position
                        words[i].SetPosition(sf::Randomizer::Random(0.f, screenWidth - words[i].GetRect().GetWidth()),
                                             sf::Randomizer::Random(0.f, screenHeight - words[i].GetRect().GetHeight()));

                        words[i].SetRotation(sf::Randomizer::Random(-80.f, 80.f));

                        words[i].SetScale(sf::Randomizer::Random(1.f, 2.f), sf::Randomizer::Random(1.f, 2.f));
                }

                // Sounds

                if (!introMusic.OpenFromFile("./rc/intro/opening.ogg"))
                {
                        // Error...
                        writeError("Couldn't load opening.ogg");
                        quitLevel = true;
                }
                introMusic.SetVolume(10.f);
                introMusic.Play();

                initializedLevel = true;
        }

        // Time to spawn a new word yet?
        if (wordDisplayTimer.GetElapsedTime() * 1000 > wordDisplayDelay
                && maxWordsShown < words.size())
        {
                // Show another word
                maxWordsShown += 1;

                // Make the next word appear a little faster
                wordDisplayDelay -= maxWordsShown;
                if (wordDisplayDelay < 0) wordDisplayDelay = 0;

                wordDisplayTimer.Reset();
        }

        // If all words have been shown, end the intro
        if (maxWordsShown >= words.size())
        {
                quitLevel = true;
        }

        // Skip the intro
        if (window.GetInput().IsKeyDown(sf::Key::Space))
        {
                quitLevel = true;
        }

        fillScreen(0, 0, 0, 0);

        // Draw words
        for (unsigned int i = 0; i < maxWordsShown; i++)
        {
                window.Draw(words[i]);
        }

        if (quitLevel)
        {
                introMusic.Stop();
        }
}

void Game::hostGame()
{
        // Incoming and outgoing data, run in a separate thread
        static sf::SocketUDP socket;

        static const unsigned short PORT = 4567;

        // Reticle drawn where the mouse is to help the player aim
        static Sprite reticle;
        static sf::Image _reticle;

        // DEBUG: a 1x1 square shape used to show coordinates
        static sf::Shape point;

        // DEBUG: test line
        static sf::Shape line;

        // Sound that plays when gun is fired
        static sf::SoundBuffer _gunSound;
        static sf::Sound gunSound;

        // Reloading sound
        static sf::SoundBuffer _reloadSound;
        static sf::Sound reloadSound;

        // Controls between shots
        static sf::Clock gunDelay;
        static const int GUN_DELAY = 400;

        // Controls the reload speed
        static sf::Clock reloadDelay;
        static const int RELOAD_DELAY = 3;

        // Sound that plays when a monster is near
        static sf::SoundBuffer _monsterSound;
        static sf::Sound monsterSound;

        // Sound of the monsters gets louder as they get nearer
        static const float MAX_MONSTER_SOUND_VOLUME = 100;
        static float monsterSoundVolume = 0.f;

        // Sound that plays when a monster attacks
        static sf::SoundBuffer _monsterAttackSound;
        static sf::Sound monsterAttackSound;

        // Music that plays in the basement
        static sf::Music basementMusic;

        // Music that plays on the roof
        static sf::Music roofMusic;

        // The higher this is, the further away we'll be able to hear the monsters from
        static const int MONSTER_SOUND_STRETCH = 100;

        // Bass note that plays on a varying interval.
        // As monsters get closer, the speed of the bass notes increases
        static sf::SoundBuffer _bassNote;
        static sf::Sound bassNote;
        static sf::Clock bassTimer;
        static const int BASS_DELAY_SLOW = 2000;
        static const int BASS_DELAY_FAST = 500;
        static const float BASS_PITCH_HIGH = 1.1f;
        static const float BASS_PITCH_LOW = 1.f;
        static const float BASS_VOLUME = 80.f;

        // Brightens the room when the player shoots
        static sf::PostFX flashEffect;

        // The higher this clock goes, the more the flash effect fades out.
        // We can simply reset it to give the flash effect life again.
        static sf::Clock flashTime;

        // The higher this is, the quicker the flash will fade out.
        static const float FLASH_TIME = 30.f;

        // Displays ammo count
        sf::String ammoDisplay;

        // Displays the floor the portal is located on
        sf::String portalFloor;

        // INITIALIZE WORLD

        if (!initializedLevel && !playedBefore)
        {
                // Try to load the flash effect
                if (flashEffect.LoadFromFile("./rc/game/colorize.sfx"))
                {
                        flashEffect.SetTexture("framebuffer", NULL);
                        flashEffect.SetParameter("color", 1.f, 1.f, 1.f);
                        flashEffect.SetParameter("intensity", 0.f);
                }
//                else
//                {
//                        writeError("Couldn't load colorize.sfx");
//                        quitLevel = true;
//                }

                // Sounds

                if (!_gunSound.LoadFromFile("./rc/game/gun.wav"))
                {
                        // Error...
                        writeError("Couldn't load gun.wav");
                        quitLevel = true;
                }
                gunSound.SetBuffer(_gunSound);
                gunSound.SetRelativeToListener(true);

                if (!_monsterSound.LoadFromFile("./rc/game/monster-sound.ogg"))
                {
                        // Error...
                        writeError("Couldn't load monster-sound.ogg");
                        quitLevel = true;
                }
                monsterSound.SetBuffer(_monsterSound);
                monsterSound.SetVolume(0.f);
                monsterSound.Play();
                monsterSound.SetLoop(true);
                monsterSound.SetRelativeToListener(true);

                if (!_bassNote.LoadFromFile("./rc/game/bass.wav"))
                {
                        // Error...
                        writeError("Couldn't load bass.wav");
                        quitLevel = true;
                }
                bassNote.SetBuffer(_bassNote);
                bassNote.SetVolume(BASS_VOLUME);
                bassNote.SetRelativeToListener(true);

                if (!_monsterAttackSound.LoadFromFile("./rc/game/monster-attack.ogg"))
                {
                        // Error...
                        writeError("Couldn't load monster-attack.ogg");
                        quitLevel = true;
                }
                monsterAttackSound.SetBuffer(_monsterAttackSound);
                monsterAttackSound.SetVolume(120.f);
                monsterAttackSound.SetRelativeToListener(true);

                if (!_reloadSound.LoadFromFile("./rc/game/reload.ogg"))
                {
                        writeError("Couldn't load reload.ogg");
                        quitLevel = true;
                }
                reloadSound.SetBuffer(_reloadSound);
                reloadSound.SetRelativeToListener(true);

                if (!basementMusic.OpenFromFile("./rc/game/hungarianrag.ogg"))
                {
                        writeError("Couldn't load hungarianrag.ogg");
                        quitLevel = true;
                }
                basementMusic.SetVolume(0.f);
                basementMusic.SetMinDistance(200.f);
                basementMusic.SetAttenuation(8.f);
                basementMusic.Play();
                basementMusic.SetPosition(rooms["Basement"]->bg.getW() + 20.f, 0.f, 0.f);

                if (!roofMusic.OpenFromFile("./rc/game/moon.ogg"))
                {
                        writeError("Couldn't load moon.ogg");
                        quitLevel = true;
                }
                roofMusic.SetVolume(0.f);
                roofMusic.SetMinDistance(200.f);
                roofMusic.SetAttenuation(1.f);
                roofMusic.Play();
                roofMusic.SetPosition(-20.f, 0.f, 0.f);

                // Test point
                point.AddPoint(0.f, 0.f, sf::Color(255, 0, 0), sf::Color(255, 0, 0));
                point.AddPoint(1.f, 0.f, sf::Color(255, 0, 0), sf::Color(255, 0, 0));
                point.AddPoint(1.f, 1.f, sf::Color(255, 0, 0), sf::Color(255, 0, 0));
                point.AddPoint(0.f, 1.f, sf::Color(255, 0, 0), sf::Color(255, 0, 0));
                point.SetOutlineWidth(2);

                // Load reticle image
                _reticle.LoadFromFile("./rc/game/reticle.png");
                reticle.SetImage(_reticle);

                // Load player images
                if (!_myPlayer.LoadFromFile("./rc/game/player1.png")
                        || !_myPlayerBArm.LoadFromFile("./rc/game/player1_barm.png")
                        || !_myPlayerFArm.LoadFromFile("./rc/game/player1_farm.png"))
                {
                        writeError("Couldn't load player 1 image(s)");
                        quitLevel = true;
                }

                myPlayer.shotsHit = 0;
                myPlayer.shotsMissed = 0;
                myPlayer.monstersKilled = 0;

                // Player is facing right by default
                myPlayer.setState(FACING_RIGHT);

                _myPlayer.SetSmooth(false);
                _myPlayerFArm.SetSmooth(false);
                _myPlayerBArm.SetSmooth(false);

                myPlayer.SetImage(_myPlayer);
                myPlayerBArm.SetImage(_myPlayerBArm);
                myPlayerFArm.SetImage(_myPlayerFArm);

                // Set rotation center of arms
                myPlayerBArm.SetCenter(ARM_ROTATION_CENTER_X, ARM_ROTATION_CENTER_Y);
                myPlayerFArm.SetCenter(ARM_ROTATION_CENTER_X, ARM_ROTATION_CENTER_Y);

                // Set collision rects
                myPlayer.setCollisionRect(52.f, 0.f, 82.f, 0.f);

                if (networkMode == MULTIPLAYER)
                {
                        if (!_theirPlayer.LoadFromFile("./rc/game/player2.png"))
                        {
                                writeError("Couldn't load player 2 image(s)");
                                quitLevel = true;
                        }

                        _theirPlayer.SetSmooth(false);

                        theirPlayer.SetImage(_theirPlayer);
                }

                // Clip the player's sub-rect
                myPlayer.SetSubRect(sf::IntRect(0, 0, PLAYER_FRAME_WIDTH, PLAYER_FRAME_HEIGHT));
                myPlayerBArm.SetSubRect(sf::IntRect(0, 0, PLAYER_FRAME_WIDTH, PLAYER_FRAME_HEIGHT));
                myPlayerFArm.SetSubRect(sf::IntRect(0, 0, PLAYER_FRAME_WIDTH, PLAYER_FRAME_HEIGHT));

                // Initialize properties of the players' animations
                myPlayer.animation.setRows(PLAYER_ROWS, myPlayer.GetImage());
                myPlayer.animation.setColumns(PLAYER_COLUMNS, myPlayer.GetImage());
                myPlayer.animation.setFirstFrame(PLAYER_FIRST_FRAME);
                myPlayer.animation.setLastFrame(PLAYER_LAST_FRAME);
                myPlayer.animation.frameDelay = PLAYER_ANIMATION_DELAY;

                // Arm animations have the same properties as the body animations
                myPlayerBArm.animation.setRows(PLAYER_ROWS, myPlayerBArm.GetImage());
                myPlayerBArm.animation.setColumns(PLAYER_COLUMNS, myPlayerBArm.GetImage());
                myPlayerBArm.animation.setFirstFrame(PLAYER_FIRST_FRAME);
                myPlayerBArm.animation.setLastFrame(PLAYER_LAST_FRAME);
                myPlayerBArm.animation.frameDelay = PLAYER_ANIMATION_DELAY;

                myPlayerFArm.animation.setRows(PLAYER_ROWS, myPlayerFArm.GetImage());
                myPlayerFArm.animation.setColumns(PLAYER_COLUMNS, myPlayerFArm.GetImage());
                myPlayerFArm.animation.setFirstFrame(PLAYER_FIRST_FRAME);
                myPlayerFArm.animation.setLastFrame(PLAYER_LAST_FRAME);
                myPlayerFArm.animation.frameDelay = PLAYER_ANIMATION_DELAY;

                myPlayer.health = PLAYER_HEALTH;
                theirPlayer.health = PLAYER_HEALTH;

                myPlayer.clipSize = 12;
                theirPlayer.clipSize = 12;

                myPlayer.ammo = PLAYER_AMMO;
                myPlayer.clip = min(myPlayer.ammo, myPlayer.clipSize);
                theirPlayer.ammo = PLAYER_AMMO;

                if (networkMode == MULTIPLAYER)
                {
                        theirPlayer.animation.setRows(PLAYER_ROWS, theirPlayer.GetImage());
                        theirPlayer.animation.setColumns(PLAYER_COLUMNS, theirPlayer.GetImage());
                        theirPlayer.animation.setFirstFrame(PLAYER_FIRST_FRAME);
                        theirPlayer.animation.setLastFrame(PLAYER_LAST_FRAME);
                        theirPlayer.animation.frameDelay = PLAYER_ANIMATION_DELAY;
                }

                // Set player 1's position
                myPlayer.SetPosition((getScreenWidth() / 2.f) - (myPlayer.getW() / 2.f), FLOOR_HEIGHT - myPlayer.getH());

                if (networkMode == MULTIPLAYER)
                {
                        theirPlayer.SetPosition(myPlayer.getX() + myPlayer.getW(), myPlayer.getY());

                        socket.Bind(PORT);

                        socket.SetBlocking(false);
                }

                initializedLevel = true;
                playedBefore = true;
        }

        // UPDATE WORLD

        // Load initial monsters and keep spawning them from portals
        spawnMonsters(initializedMonsters);
        initializedMonsters = true;

        // Window input
        const sf::Input& input = window.GetInput();
        float deltaTime = window.GetFrameTime();

        // Spinning effect for portal
        portal.rotate(deltaTime);

        // Set reticle to mouse position (adjusted for camera movement)
        reticle.SetPosition(camera.GetRect().Left + input.GetMouseX() - (reticle.getW() / 2.f), input.GetMouseY() - (reticle.getH() / 2.f));

        // Handle the player controls, but pass the direction we're aiming in
        // (Note that the player may be moving in the opposite direction they're aiming,
        // and we need to handle that appropriately)
        if (reticle.getCenterX() < myPlayer.getCenterX())
        {
                handlePlayerControls(FACING_LEFT);
        }
        else
        {
                handlePlayerControls(FACING_RIGHT);
        }

        myPlayer.moveAndSave(deltaTime);

        // The player is the "listener," and all sounds will be played relative to them
        sf::Listener::SetPosition(myPlayer.getX(), 0.f, 0.f);

        moveAndAnimateMonsters(deltaTime);

        sf::Socket::Status sendStatus;
        sf::Socket::Status rcvStatus;
        if (networkMode == MULTIPLAYER)
        {
                if (socket.IsValid())
                {
                        // SEND position data to the client player
                        sf::Packet outPacket;
                        outPacket << myPlayer.getX() << myPlayer.getY();
                        // Note: Send() and Receive() seem to modify the IP you provide to be your own.
                        // We don't want this to change our permanent copy of the client's IP, so send a copy.
                        sf::IPAddress tip = theirIP;
                        sendStatus = socket.Send(outPacket, tip, PORT);

                        // RECEIVE packets
                        float x = theirPlayer.getX();
                        float y = theirPlayer.getY();
                        sf::Packet inPacket;
                        unsigned short port = PORT;
                        tip = theirIP;
                        rcvStatus = socket.Receive(inPacket, tip, port);

                        if (rcvStatus == sf::Socket::Done)
                        {
                                if (!(inPacket >> x >> y))
                                {
                                        // Error...
                                }
                                else
                                {
                                        theirPlayer.SetPosition(x, y);
                                }
                        }
                }
                else
                {
                        // Error!
                        // TODO: graceful quit and error message
                        writeError("Invalid socket");
                        quitLevel = true;
                }
        }

        // COLLISION DETECTION

        // Check if the player is touching a pickup
        for (unsigned int i = 0; i < pickups.size(); i++)
        {
                if (pickups[i] != NULL)
                {
                        if (myPlayer.parentRoom == pickups[i]->parentRoom
                                && myPlayer.getX() + myPlayer.getCollisionRect().Left >= pickups[i]->getX()
                                && myPlayer.getX() + myPlayer.getW() - myPlayer.getCollisionRect().Right <= pickups[i]->getX() + pickups[i]->getW()
                                && input.IsKeyDown(sf::Key::Space))
                        {
                                // Remove the pickup
                                delete pickups[i];
                                pickups[i] = NULL;

                                myPlayer.health = PLAYER_HEALTH;
                                myPlayer.ammo += AMMO_PICKUP_INCREASE;

                                // Play a sound
                                reloadSound.Play();
                        }
                }
        }

        // Check if the player is colliding with the walls
        if (myPlayer.parentRoom->hasLeftWall && myPlayer.getX() < myPlayer.parentRoom->leftBoundary)
        {
                myPlayer.SetPosition(myPlayer.parentRoom->leftBoundary, myPlayer.getY());
        }
        else if (myPlayer.parentRoom->hasRightWall && myPlayer.getX() + myPlayer.getW() > myPlayer.parentRoom->bg.getW() - myPlayer.parentRoom->rightBoundary)
        {
                myPlayer.SetPosition(myPlayer.parentRoom->bg.getW() - myPlayer.parentRoom->rightBoundary - myPlayer.getW(), myPlayer.getY());
        }

        // Player leaving right side of room?
        if (myPlayer.getX() > myPlayer.parentRoom->bg.getW())
        {
                // If there's a connected room on the right
                if (myPlayer.parentRoom->links[RIGHT] != NULL)
                {
                        // Move into the left side of that room
                        myPlayer.parentRoom = myPlayer.parentRoom->links[RIGHT];

                        myPlayer.SetPosition(-myPlayer.getW() / 2.f, myPlayer.getY());
                }
        }
        // Player leaving left side of room?
        else if (myPlayer.getX() < -myPlayer.getW())
        {
                // If there's a connected room on the left
                if (myPlayer.parentRoom->links[LEFT] != NULL)
                {
                        // Move into the left side of that room
                        myPlayer.parentRoom = myPlayer.parentRoom->links[LEFT];

                        myPlayer.SetPosition(myPlayer.parentRoom->bg.getW() - (myPlayer.getW() / 2.f), myPlayer.getY());
                }
        }
        // Player moving down?
        else if (input.IsKeyDown(sf::Key::S))
        {
                // If there's a connected room on the bottom and we're standing in the stairwell
                if (myPlayer.parentRoom->links[DOWN] != NULL
                        && myPlayer.getX() >= myPlayer.parentRoom->stairsDown.Left
                        && myPlayer.getX() + myPlayer.getW() <= myPlayer.parentRoom->stairsDown.Right)
                {
                        // Move below the room
                        myPlayer.parentRoom = myPlayer.parentRoom->links[DOWN];

                        // Move into the center of this room's UP stairwell
                        myPlayer.SetPosition(((myPlayer.parentRoom->stairsUp.Right - myPlayer.parentRoom->stairsUp.Left) / 2.f) + myPlayer.parentRoom->stairsUp.Left - (myPlayer.getW() / 2.f),
                                                myPlayer.getY());
                }
        }
        // Player moving up?
        else if (input.IsKeyDown(sf::Key::W))
        {
                // If there's a connected room on the top and we're standing in the stairwell
                if (myPlayer.parentRoom->links[UP] != NULL
                        && myPlayer.getX() >= myPlayer.parentRoom->stairsUp.Left
                        && myPlayer.getX() + myPlayer.getW() <= myPlayer.parentRoom->stairsUp.Right)
                {
                        // Move above the room
                        myPlayer.parentRoom = myPlayer.parentRoom->links[UP];

                        // Move into the center of this room's DOWN stairwell
                        myPlayer.SetPosition(((myPlayer.parentRoom->stairsDown.Right - myPlayer.parentRoom->stairsDown.Left) / 2.f) + myPlayer.parentRoom->stairsDown.Left - (myPlayer.getW() / 2.f),
                                                myPlayer.getY());
                }
        }

        // Monster CD
        monsterAI();

        // Have we changed directions?
        if (myPlayer.getChangedState())
        {
                if (myPlayer.getState() == FACING_LEFT)
                {
                        // Set rotation center of arms
                        myPlayerBArm.SetCenter(myPlayerBArm.getW() - ARM_ROTATION_CENTER_X, ARM_ROTATION_CENTER_Y);
                        myPlayerFArm.SetCenter(myPlayerFArm.getW() - ARM_ROTATION_CENTER_X, ARM_ROTATION_CENTER_Y);
                }
                else if (myPlayer.getState() == FACING_RIGHT)
                {
                        // Set rotation center of arms
                        myPlayerBArm.SetCenter(ARM_ROTATION_CENTER_X, ARM_ROTATION_CENTER_Y);
                        myPlayerFArm.SetCenter(ARM_ROTATION_CENTER_X, ARM_ROTATION_CENTER_Y);
                }
        }

        // Is the player being hurt by a monster?
        for (unsigned int i = 0; i < monsters.size() && !quitLevel; i++)
        {
                if (monsters[i] != NULL && myPlayer.parentRoom == monsters[i]->parentRoom)
                {
                        // if we're not NOT colliding
                        if (!(myPlayer.getX() + myPlayer.getCollisionRect().Left > monsters[i]->getX() + monsters[i]->getW())
                                && !(myPlayer.getX() + myPlayer.getW() - myPlayer.getCollisionRect().Right < monsters[i]->getX()))
                        {
                                if (monsters[i]->attackTimer.GetElapsedTime() * 1000 > monsters[i]->attackDelay)
                                {
                                        myPlayer.health--;
                                        
                                        monsters[i]->attackTimer.Reset();

                                        if (monsterAttackSound.GetStatus() != sf::Sound::Playing)
                                                monsterAttackSound.Play();
                                }
                                
                                // Is the player dead?
                                if (myPlayer.health <= 0)
                                {
                                        myPlayer.health = 0;

                                        // We lost the game!
                                        gameWon = false;

                                        // Show the game over screen
                                        gameStates.push(&Game::gameOver);

                                        gameWon = false;

                                        quitLevel = true;
                                }
                        }
                }
        }

        // Move the player's arms to keep up with their body
        // NOTE: this is done after collision detection so the arms don't detach
        // when the player does collision response
        if (myPlayer.getState() == FACING_RIGHT)
        {
                myPlayerBArm.SetPosition(myPlayer.getX() + ARM_ROTATION_CENTER_X, myPlayer.getY() + ARM_ROTATION_CENTER_Y);
                myPlayerFArm.SetPosition(myPlayer.getX() + ARM_ROTATION_CENTER_X, myPlayer.getY() + ARM_ROTATION_CENTER_Y);
        }
        else if (myPlayer.getState() == FACING_LEFT)
        {
                myPlayerBArm.SetPosition(myPlayer.getX() + (myPlayer.getW() - ARM_ROTATION_CENTER_X), myPlayer.getY() + ARM_ROTATION_CENTER_Y);
                myPlayerFArm.SetPosition(myPlayer.getX() + (myPlayer.getW() - ARM_ROTATION_CENTER_X), myPlayer.getY() + ARM_ROTATION_CENTER_Y);
        }

        // Camera scrolling
        handleCameras(myPlayer.parentRoom->bg.getW());

        // Animate players
        if (myPlayer.xMov != 0.f)
        {
                myPlayer.animation.animate(myPlayer);
                myPlayerBArm.animation.animate(myPlayerBArm);
                myPlayerFArm.animation.animate(myPlayerFArm);
        }
        else
        {
                // Stand still
                myPlayer.animation.setFrame(1, myPlayer);
                myPlayerBArm.animation.setFrame(1, myPlayer);
                myPlayerFArm.animation.setFrame(1, myPlayer);
        }

        if (networkMode == MULTIPLAYER && theirPlayer.xMov != 0.f)
        {
                theirPlayer.animation.animate(theirPlayer);
        }
        else if (networkMode == MULTIPLAYER)
        {
                // Stand still
                theirPlayer.animation.setFrame(1, myPlayer);
        }

        // Rotate the arms based on the position of the mouse

        float shootingAngle = 0.f;
                                
        // Don't divide by 0!
        if (myPlayer.getState() == FACING_RIGHT && reticle.getCenterX() >= myPlayer.getCenterX())
        {
                // Distance between arm center and mouse pointer
                float d = distance(reticle.getCenterX(), reticle.getCenterY(), myPlayerFArm.getX(), myPlayerFArm.getY());

                // Don't divide by 0!
                if (d != 0)
                {
                        shootingAngle = std::asin((myPlayerFArm.getY() - reticle.getCenterY()) / d) * RADIANS_TO_DEGREES;

                        // Don't exceed the min/max angles
                        if (shootingAngle > MAX_AIM_ANGLE) shootingAngle = MAX_AIM_ANGLE;
                        else if (shootingAngle < MIN_AIM_ANGLE) shootingAngle = MIN_AIM_ANGLE;

                        myPlayerBArm.SetRotation(shootingAngle);
                        myPlayerFArm.SetRotation(shootingAngle);
                }
        }
        else if (myPlayer.getState() == FACING_LEFT && reticle.getCenterX() < myPlayer.getCenterX())
        {
                // Distance between arm center and mouse pointer
                float d = distance(reticle.getCenterX(), reticle.getCenterY(), myPlayerFArm.getX(), myPlayerFArm.getY());

                // Don't divide by 0!
                if (d != 0)
                {
                        shootingAngle = std::asin((myPlayerFArm.getY() - reticle.getCenterY()) / d) * 100.f;

                        // Don't exceed the min/max angles
                        if (shootingAngle > MAX_AIM_ANGLE) shootingAngle = MAX_AIM_ANGLE;
                        else if (shootingAngle < MIN_AIM_ANGLE) shootingAngle = MIN_AIM_ANGLE;

                        myPlayerBArm.SetRotation(-shootingAngle);
                        myPlayerFArm.SetRotation(-shootingAngle);
                }
        }

        // If we're in the basement, play the rag song in the background
        if (myPlayer.parentRoom == rooms["Basement"])
        {
                basementMusic.SetVolume(100.f);
        }
        else
        {
                basementMusic.SetVolume(0.f);
        }

        // If we're on the roof, play the ukelele song
        if (myPlayer.parentRoom == rooms["Roof"])
        {
                roofMusic.SetVolume(100.f);
        }
        else
        {
                roofMusic.SetVolume(0.f);
        }

        // Change sound volume depending on distance from the nearest monster
        float nearest = -1;
        for (unsigned int i = 0; i < monsters.size(); i++)
        {
                if (monsters[i] != NULL && myPlayer.parentRoom == monsters[i]->parentRoom)
                {
                        // Distance from monster i
                        float distance = std::abs(myPlayer.getX() - monsters[i]->getX());

                        // Initial comparison basis
                        if (nearest == -1)
                                nearest = distance;

                        if (distance < nearest)
                                nearest = distance;
                }
        }

        // The farther the monsters are, the quieter the sound should be

        // Did we find a nearest monster in the room?
        if (nearest != -1)
        {
                // Don't divide by 0!
                nearest /= MONSTER_SOUND_STRETCH;
                if (nearest < 1)
                        nearest = 1;
                monsterSoundVolume = MAX_MONSTER_SOUND_VOLUME * (1 / nearest);
                monsterSound.SetVolume(monsterSoundVolume);

                // Play the fast-tempo bass notes
                if (bassTimer.GetElapsedTime() * 1000 > BASS_DELAY_FAST)
                {
                        bassNote.Play();
                        bassTimer.Reset();
                        
                        // DEBUG
                        bassNote.SetPitch(BASS_PITCH_HIGH);
                }
        }
        // No monsters in this room, so turn off the sounds
        else
        {
                monsterSoundVolume = 0.f;
                monsterSound.SetVolume(monsterSoundVolume);

                // Play the slow-tempo bass notes
                if (bassTimer.GetElapsedTime() * 1000 > BASS_DELAY_SLOW)
                {
                        bassNote.Play();
                        bassTimer.Reset();

                        // Debug
                        bassNote.SetPitch(BASS_PITCH_LOW);
                }
        }

        // Reload the gun manually
        if (input.IsKeyDown(sf::Key::R) && myPlayer.clip < myPlayer.clipSize)
        {
                reloadSound.Play();

                myPlayer.clip = min(myPlayer.ammo, myPlayer.clipSize);

                reloadDelay.Reset();
        }

        // Fire the gun
        if (myPlayer.ammo > 0 && reloadDelay.GetElapsedTime() > RELOAD_DELAY && gunDelay.GetElapsedTime() * 1000 > GUN_DELAY && input.IsMouseButtonDown(sf::Mouse::Left))
        {
                gunSound.Play();

                // Start the flash again
                flashTime.Reset();

                // Decrease ammo
                myPlayer.clip--;
                myPlayer.ammo--;

                // Reload if the player's out of ammo or if they're pressing R
                if (myPlayer.clip <= 0)
                {
                        reloadSound.Play();

                        myPlayer.clip = min(myPlayer.ammo, myPlayer.clipSize);

                        reloadDelay.Reset();
                }

                // Rotate the arms a bit to simulate recoil
                if (myPlayer.getState() == FACING_RIGHT)
                {
                        myPlayerFArm.Rotate(RECOIL_ANGLE);
                        myPlayerBArm.Rotate(RECOIL_ANGLE);
                }
                else if (myPlayer.getState() == FACING_LEFT)
                {
                        myPlayerFArm.Rotate(-RECOIL_ANGLE);
                        myPlayerBArm.Rotate(-RECOIL_ANGLE);
                }

                bool foundTarget = false;
                float nearest = 0;
                int nearestMonster = -1;
                bool hitMonster = false;

                // TODO: make this only shoot monsters in the line of fire
                for (unsigned int i = 0; i < monsters.size(); i++)
                {
                        // Only shoot the monster in front of you, and ignore monsters behind you
                        if (monsters[i] != NULL && monsters[i]->parentRoom == myPlayer.parentRoom
                                && ((myPlayer.getState() == FACING_RIGHT && monsters[i]->getX() > myPlayer.getX() + ARM_ROTATION_CENTER_X)
                                     || (myPlayer.getState() == FACING_LEFT && monsters[i]->getX() < myPlayer.getX() + myPlayer.getW() - ARM_ROTATION_CENTER_X)))

                        {
                                float distance;

                                // Distance from monster i
                                if (myPlayer.getState() == FACING_RIGHT)
                                {
                                        distance = monsters[i]->getX() - myPlayer.getX() + ARM_ROTATION_CENTER_X;
                                }
                                else if (myPlayer.getState() == FACING_LEFT)
                                {
                                        distance = myPlayer.getX() + myPlayer.getW() - ARM_ROTATION_CENTER_X - monsters[i]->getX() + monsters[i]->getW();
                                }

                                // Initial comparison basis
                                if (!foundTarget || distance < nearest)
                                {
                                        nearest = distance;

                                        nearestMonster = i;

                                        foundTarget = true;
                                }
                        }
                }

                // Nearest is now 0 (no monsters in the line of fire), or the distance between the player's arm
                // and the nearest side of the nearest monster

               if (foundTarget && nearestMonster != -1)
               {
                        float m = std::tan(myPlayerFArm.GetRotation() / RADIANS_TO_DEGREES);

                        if (myPlayer.getState() == FACING_LEFT)
                                m *= -1;

                        float x = nearest;

                        float magicFactor;
                        if (myPlayer.getState() == FACING_RIGHT)
                        {
                                magicFactor = 58.29f;
                                line = sf::Shape::Line(myPlayer.getX() + ARM_ROTATION_CENTER_X,
                                                       myPlayer.getY() + ARM_ROTATION_CENTER_Y,
                                                       myPlayer.getX() + ARM_ROTATION_CENTER_X + nearest,
                                                       getScreenHeight() - (m * x) - myPlayer.getY() - ARM_ROTATION_CENTER_Y + magicFactor,  // y = mx+b
                                                       1.f, sf::Color(255, 0, 0), 0.f, sf::Color(255, 0, 0));
                        }
                        else if (myPlayer.getState() == FACING_LEFT)
                        {
                                magicFactor = 48.29f;
                                line = sf::Shape::Line(myPlayer.getX() + myPlayer.getW() - ARM_ROTATION_CENTER_X - nearest,
                                                       getScreenHeight() - (m * x) - myPlayer.getY() - ARM_ROTATION_CENTER_Y + magicFactor,  // y = mx+b
                                                       myPlayer.getX() + myPlayer.getW() - ARM_ROTATION_CENTER_X,
                                                       myPlayer.getY() + ARM_ROTATION_CENTER_Y,
                                                       1.f, sf::Color(255, 0, 0), 0.f, sf::Color(255, 0, 0));
                        }

                        // Using y = mx + b, we found the point on the axis of fire with the same x-coordinate as the nearest monster's nearest side.
                        // If this point is above or below that nearest side, then there is no bullet collision.
                        float hitY = getScreenHeight() - (m * std::abs(x)) - myPlayer.getY() - ARM_ROTATION_CENTER_Y + magicFactor;

                        if (hitY >= monsters[nearestMonster]->getY() && hitY <= monsters[nearestMonster]->getY() + monsters[nearestMonster]->getH())
                        {
                                // Hurt the monster
                                monsters[nearestMonster]->hurt(PLAYER_DAMAGE);

                                // Record the hit
                                myPlayer.shotsHit++;

                                // If we killed the monster
                                if (monsters[nearestMonster]->getState() == DEAD)
                                {
                                        // Delete this entry in the table
                                        delete monsters[nearestMonster];
                                        monsters[nearestMonster] = NULL;

                                        livingMonsters--;

                                        myPlayer.monstersKilled++;
                                }

                                hitMonster = true;
                        }
                        else
                        {
                                // We missed the monster
                                myPlayer.shotsMissed++;
                        }
                }

                if (!hitMonster && myPlayer.parentRoom == portal.parentRoom
                                && ((myPlayer.getState() == FACING_RIGHT && portal.getX() > myPlayer.getX() + ARM_ROTATION_CENTER_X)
                                     || (myPlayer.getState() == FACING_LEFT && portal.getX() < myPlayer.getX() + myPlayer.getW() - ARM_ROTATION_CENTER_X)))
                {
                        // TODO: only make this hit the portal if it's in the line of fire
                        portal.health--;

                        myPlayer.shotsHit++;

                        // Make the portal's color reflect their health
                        float scaleFactor = (255.f / portal.maxHealth);
                        // Approaches 0 as health approaches maxHealth
                        sf::Uint8 colorDrain = 255 - ((portal.maxHealth - portal.health) * (255 / portal.maxHealth));
                        portal.SetColor(sf::Color(colorDrain, colorDrain, colorDrain));

                        // Did we destroy the portal?
                        if (portal.health <= 0)
                        {
                                portal.lives--;

                                // Portal is'nt dead yet, so relocate it
                                if (portal.lives > 0)
                                {
                                        portal.health = portal.maxHealth;
                                        
                                        // Relocate the portal

                                        // Pick a random room to put the portal in,
                                        // repeat until we have a random room that no player is in
                                        std::map<std::string, Room*>::iterator it;
                                        do
                                        {
                                                it = rooms.begin();

                                                int randRoom = sf::Randomizer::Random(0, rooms.size() - 1);

                                                // Iterate through the map of rooms until we have the randomly selected room
                                                for (int i = 0; i < randRoom && it != rooms.end(); i++)
                                                { 
                                                        it++;
                                                }
                                        } while (it->second == myPlayer.parentRoom);

                                        // Place the portal in the center of the room
                                        portal.SetPosition(it->second->bg.getW() / 2.f, portal.getY());

                                        portal.parentRoom = it->second;

                                        // Flash
                                        flashTime.Reset();

                                        // Replenish the player's health
                                        myPlayer.health = PLAYER_HEALTH;
                                }
                                else
                                {
                                        // Killed the portal!  Game over
                                        gameStates.push(&Game::finalCutscene);
                                        quitLevel = true;
                                }
                        }
                }

                gunDelay.Reset();

        }

        float intensity = 0.f;
        float elapsedTime = flashTime.GetElapsedTime();
        if (elapsedTime != 0)
        {
                intensity = 1.f / (flashTime.GetElapsedTime() * FLASH_TIME);
        }
        if (intensity > 1.f) intensity = 1.f;
        flashEffect.SetParameter("intensity", intensity);

        // Make the player's color reflect their health
        float scaleFactor = (255.f / PLAYER_HEALTH);
        // Approaches 0 as health approaches MAX_HEALTH
        sf::Uint8 colorDrain = 255 - ((PLAYER_HEALTH - myPlayer.health) * (255 / PLAYER_HEALTH));
        myPlayer.SetColor(sf::Color(255, colorDrain, colorDrain));
        myPlayerBArm.SetColor(sf::Color(255, colorDrain, colorDrain));
        myPlayerFArm.SetColor(sf::Color(255, colorDrain, colorDrain));

        // DEBUG
//        // Draw a line through the axis of aiming
//        float m = std::tan(myPlayerBArm.GetRotation() / RADIANS_TO_DEGREES);
////        float x = 200.f;
//        float x = 500.f;
//        // Translation factor/y-intercept (FIXME: the 34.39 is a magic number that needs account for, so I'm obviously missing something)
//        line = sf::Shape::Line(myPlayer.getX() + ARM_ROTATION_CENTER_X,
//                               myPlayer.getY() + ARM_ROTATION_CENTER_Y,
//                               myPlayer.getX() + ARM_ROTATION_CENTER_X + x,
//                               getScreenHeight() - (m * x) - myPlayer.getY() - ARM_ROTATION_CENTER_Y + 34.29,  // y = mx+b
//                               1.f, sf::Color(255, 0, 0), 0.f, sf::Color(255, 0, 0));


        // Update the display of the player's ammo
        std::stringstream ss;
        ss << "Clip: " << myPlayer.clip << "\n";
        ss << "Ammo: " << myPlayer.ammo;
        ammoDisplay.SetText(ss.str());
        ammoDisplay.SetPosition(20.f + camera.GetRect().Left, getScreenHeight() - ammoDisplay.GetRect().GetHeight() - 20.f);

        // Tell the player what floor to find the portal on
        if (portal.parentRoom->filenames[0].substr(0, 2) == myPlayer.parentRoom->filenames[0].substr(0, 2))
        {
                portalFloor.SetText("Portal is on this floor!");
        }
        else
        {
                portalFloor.SetText("");
        }

        portalFloor.SetPosition(20.f + camera.GetRect().Left, 20.f);

        // DISPLAY WORLD

        // Backgrounds (except doorframe slices, if there are any)
        for (unsigned int i = 0; i < myPlayer.parentRoom->bg.objects.size(); i++)
        {
                if (!(myPlayer.parentRoom->hasLeftDoor && i == 0) && !(myPlayer.parentRoom->hasRightDoor && i == myPlayer.parentRoom->bg.objects.size() - 1))
                {
                        window.Draw(*myPlayer.parentRoom->bg.getSprite(i));
                }
        }

        if (networkMode == MULTIPLAYER)
        {
                window.Draw(theirPlayer);
        }
                
        window.Draw(myPlayerBArm);
        window.Draw(myPlayer);
        window.Draw(myPlayerFArm);

        // Pickups
        for (unsigned int i = 0; i < pickups.size(); i++)
        {
                if (pickups[i] != NULL && pickups[i]->parentRoom == myPlayer.parentRoom)
                {
                        window.Draw(*pickups[i]);
                }
        }

        // Draw monsters
        for (unsigned int i = 0; i < monsters.size(); i++)
        {
                if (monsters[i] != NULL)
                {
                        // Don't draw any monsters that aren't in our room
                        if (myPlayer.parentRoom == monsters[i]->parentRoom)
                        {
                                window.Draw(*monsters[i]);
                        }
                }
        }

        // If we're in the portal room, draw it
        if (myPlayer.parentRoom == portal.parentRoom)
        {
                window.Draw(portal);
        }

        if (myPlayer.parentRoom->hasLeftDoor)
        {
                // Draw left doorframe on top of player
                window.Draw(*myPlayer.parentRoom->bg.getSprite(0));
        }
        if (myPlayer.parentRoom->hasRightDoor)
        {
                // Draw right doorframe on top of player
                window.Draw(*myPlayer.parentRoom->bg.getSprite(myPlayer.parentRoom->bg.objects.size() - 1));
        }

        // Post effects
        window.Draw(flashEffect);

        window.Draw(reticle);

        window.Draw(ammoDisplay);

        window.Draw(portalFloor);

        if (debugOn)
        {
                printFramerate();

                if (networkMode == MULTIPLAYER)
                {
                        if (sendStatus == sf::Socket::Done)
                        {
                                printText(20.f, 20.f, "Done");
                        }
                        else if (sendStatus == sf::Socket::NotReady)
                        {
                                printText(20.f, 20.f, "NotReady");
                        }
                        else if (sendStatus == sf::Socket::Disconnected)
                        {
                                printText(20.f, 20.f, "Disconnected");
                        }
                        else if (sendStatus == sf::Socket::Error)
                        {
                                printText(20.f, 20.f, "Error");
                        }

                        if (rcvStatus == sf::Socket::Done)
                        {
                                printText(20.f, 40.f, "Done");
                        }
                        else if (rcvStatus == sf::Socket::NotReady)
                        {
                                printText(20.f, 40.f, "NotReady");
                        }
                        else if (rcvStatus == sf::Socket::Disconnected)
                        {
                                printText(20.f, 40.f, "Disconnected");
                        }
                        else if (rcvStatus == sf::Socket::Error)
                        {
                                printText(20.f, 40.f, "Error");
                        }
                }

//                printText(20.f, 100.f, theirIP.ToString());

//                for (unsigned int i = 0; i < monsters.size(); i++)
//                {
//                        if (monsters[i] != NULL)
//                        {
//                                std::stringstream ss;
//                                ss << "in " << monsters[i]->parentRoom->filenames[0] << " at " << monsters[i]->getX() << " going " << monsters[i]->xMov;
//                                printText(20.f + camera.GetRect().Left, 30.f * i + 20.f, ss.str());
//                        }
//                }

//                std::stringstream ss;
////                ss << myPlayerBArm.GetRotation();
//                ss << (myPlayer.getY() + ARM_ROTATION_CENTER_Y) - line.GetPointPosition(1).y;
//                printText(20.f, getScreenHeight() - 40.f, ss.str());

//                window.Draw(point);

//                window.Draw(line);
        }

        // If a quit message was posted, close the socket
        if (quitLevel)
        {
                monsterSound.Stop();
                monsterAttackSound.Stop();
                bassNote.Stop();
                basementMusic.Stop();
                roofMusic.Stop();
                reloadSound.Stop();

                if (networkMode == MULTIPLAYER)
                {
                        socket.Close();
                }

                resetCamera();
        }
}

void Game::clientGame()
{
        static sf::SocketUDP socket;

        static const unsigned short PORT = 4567;

        static BigSprite bg;
        // A mask of shadows that gives the room some artificial lighting
        static BigSprite bgMask;

        // INITIALIZE WORLD

        if (!initializedLevel)
        {
                if (!bg.addSprite("./rc/game/2f_candlehallway_bg_1.png")
                        || !bg.addSprite("./rc/game/2f_candlehallway_bg_2.png"))
                {
                        writeError("Couldn't load 2f_candlehallway_bg");
                        quitLevel = true;
                }

                if (!bgMask.addSprite("./rc/game/2f_candlehallway_shadows_1.png")
                        || !bgMask.addSprite("./rc/game/2f_candlehallway_shadows_2.png"))
                {
                        writeError("Couldn't load 2f_candlehallway_shadows");
                        quitLevel = true;
                }

                // Load player images
                if (!_myPlayer.LoadFromFile("./rc/game/player2.png"))
                {
                        writeError("Couldn't load player2.png");
                        quitLevel = true;
                }
                myPlayer.SetImage(_myPlayer);

                if (!_theirPlayer.LoadFromFile("./rc/game/player1.png"))
                {
                        writeError("Couldn't load player1.png");
                        quitLevel = true;
                }
                theirPlayer.SetImage(_theirPlayer);

                // Set player 1's position
                myPlayer.SetPosition((getScreenWidth() / 2.f) - (myPlayer.getW() / 2.f), 270.f);

                theirPlayer.SetPosition(myPlayer.getX() + myPlayer.getW(), myPlayer.getY());

                // Listen to this port
                socket.Bind(PORT);

                socket.SetBlocking(false);

                initializedLevel = true;
        }

        // UPDATE WORLD

        float deltaTime = window.GetFrameTime();

        // handlePlayerControls(....)

        myPlayer.moveAndSave(deltaTime);

        sf::Socket::Status sendStatus;
        sf::Socket::Status rcvStatus;
        if (socket.IsValid())
        {
                // RECEIVE packets
                float x = theirPlayer.getX();
                float y = theirPlayer.getY();
                sf::Packet inPacket;
                unsigned short port = PORT;
                // Note: Send() and Receive() seem to modify the IP address you send to to be your own,
                // so we don't want to send our permanent copy of the host's address.
                sf::IPAddress tip = theirIP;
                rcvStatus = socket.Receive(inPacket, tip, port);

                if (rcvStatus == sf::Socket::Done)
                {
                        if (!(inPacket >> x >> y))
                        {
                                // Error...
                        }
                        else
                        {
                                theirPlayer.SetPosition(x, y);
                        }
                }

                // SEND position data to the host player
                sf::Packet outPacket;
                outPacket << myPlayer.getX() << myPlayer.getY();
                tip = theirIP;
                sendStatus = socket.Send(outPacket, tip, PORT);
        }
        else
        {
                // Error!
                // TODO: graceful quit and error message
                writeError("Invalid socket");
                quitLevel = true;
        }

        handleCameras(bg.getW());

        // DISPLAY WORLD

        for (unsigned int i = 0; i < bg.objects.size(); i++)
        {
                window.Draw(*bg.getSprite(i));
        }

        if (networkMode == MULTIPLAYER)
        {
                window.Draw(theirPlayer);
        }

        window.Draw(theirPlayer);
        window.Draw(myPlayer);

        // Draw mask of shadows over the entire room
        for (unsigned int i = 0; i < bgMask.objects.size(); i++)
        {
                window.Draw(*bgMask.getSprite(i));
        }

        if (debugOn)
        {
                printFramerate();

                if (sendStatus == sf::Socket::Done)
                {
                        printText(20.f, 20.f, "Done");
                }
                else if (sendStatus == sf::Socket::NotReady)
                {
                        printText(20.f, 20.f, "NotReady");
                }
                else if (sendStatus == sf::Socket::Disconnected)
                {
                        printText(20.f, 20.f, "Disconnected");
                }
                else if (sendStatus == sf::Socket::Error)
                {
                        printText(20.f, 20.f, "Error");
                }

                if (rcvStatus == sf::Socket::Done)
                {
                        printText(20.f, 40.f, "Done");
                }
                else if (rcvStatus == sf::Socket::NotReady)
                {
                        printText(20.f, 40.f, "NotReady");
                }
                else if (rcvStatus == sf::Socket::Disconnected)
                {
                        printText(20.f, 40.f, "Disconnected");
                }
                else if (rcvStatus == sf::Socket::Error)
                {
                        printText(20.f, 40.f, "Error");
                }

                printText(20.f, 100.f, theirIP.ToString());
        }

        // If a quit message was posted, close the socket
        if (quitLevel)
        {
                socket.Close();
        }
}

void Game::gameOver()
{
        // Background if the player lost the game
        static sf::Image _gameOverBG;
        static Sprite gameOverBG;

        // Displays the player's accuracy and monsters killed
        static sf::String statsStr;

        if (!initializedLevel)
        {
                if (gameWon)
                {
                        if (!_gameOverBG.LoadFromFile("./rc/game-over/win.png"))
                        {
                                // Error...
                                writeError("Couldn't lose lose.png");
                                quitLevel = true;
                        }
                        _gameOverBG.SetSmooth(false);
                }
                else
                {
                        if (!_gameOverBG.LoadFromFile("./rc/game-over/lose.png"))
                        {
                                // Error...
                                writeError("Couldn't lose lose.png");
                                quitLevel = true;
                        }
                        _gameOverBG.SetSmooth(false);
                }

                gameOverBG.SetImage(_gameOverBG);

                float totalShots = float(myPlayer.shotsHit) + float(myPlayer.shotsMissed);
                float accuracy;
                accuracy = ((totalShots > 0) ? (myPlayer.shotsHit / totalShots) : 0) * 100;
                std::stringstream ss;
                ss << "Accuracy: " << accuracy << "%\n"
                   << "Monsters killed: " << myPlayer.monstersKilled;
                statsStr.SetText(ss.str());
                statsStr.SetPosition((getScreenWidth() / 2.f) - (statsStr.GetRect().GetWidth() / 2.f), getScreenHeight() - statsStr.GetRect().GetHeight() - 40.f);

                initializedLevel = true;
        }

        // UPDATE

        // Does the player want to start again?
        if (window.GetInput().IsKeyDown(sf::Key::Return))
        {
                // Go back to the game selection screen
                gameStates.push(&Game::gameSelection);


                quitLevel = true;
        }

        // DRAW
        window.Draw(gameOverBG);

        window.Draw(statsStr);
}

void Game::finalCutscene()
{
        static const float PORTAL_SCALE_STEP = 0.01f;
        static const float MAX_PORTAL_SCALE = 4.f;
        static float portalScale = 1.f;
        static sf::Clock portalScaleDelay;
        static const int PORTAL_SCALE_DELAY = 10;

        if (!initializedLevel)
        {
                // Have to move left to get to portal
                if (portal.getX() < myPlayer.getCenterX())
                {
                        myPlayer.xMov = -RUN_SPEED;
                }
                // Have to move right to get to the portal
                else
                {
                        myPlayer.xMov = RUN_SPEED;
                }

                // Walk forward animation
                myPlayer.animation.setPosDirection();

                initializedLevel = true;
        }

        // UPDATE WORLD

        float deltaTime = window.GetFrameTime();

        // Spinning effect for portal
        portal.rotate(deltaTime);

        myPlayer.moveAndSave(deltaTime);

        // COLLISION DETECTION

        // Are we inside the portal?
        if (!(myPlayer.getCenterX() < portal.getX() - 20.f)
                && !(myPlayer.getCenterX() > portal.getX() + 20.f))
        {
                // Freeze the player
                myPlayer.xMov = 0.f;

                // Start making the portal bigger

                // Are we done scaling yet?
                // Scale up, if it's been long enough
                if (portalScaleDelay.GetElapsedTime() * 1000 > PORTAL_SCALE_DELAY)
                {
                        portalScale += PORTAL_SCALE_STEP;

                        portal.SetScale(portalScale, portalScale);

                        portalScaleDelay.Reset();

                        // Have we scaled up enough?
                        if (portalScale >= MAX_PORTAL_SCALE)
                        {
                                // We won the game!
                                gameWon = true;

                                gameStates.push(&Game::gameOver);

                                // End the game
                                quitLevel = true;
                        }
                }
        }

        // Move the player's arms to keep up with their body
        if (myPlayer.getState() == FACING_RIGHT)
        {
                myPlayerBArm.SetPosition(myPlayer.getX() + ARM_ROTATION_CENTER_X, myPlayer.getY() + ARM_ROTATION_CENTER_Y);
                myPlayerFArm.SetPosition(myPlayer.getX() + ARM_ROTATION_CENTER_X, myPlayer.getY() + ARM_ROTATION_CENTER_Y);
        }
        else if (myPlayer.getState() == FACING_LEFT)
        {
                myPlayerBArm.SetPosition(myPlayer.getX() + (myPlayer.getW() - ARM_ROTATION_CENTER_X), myPlayer.getY() + ARM_ROTATION_CENTER_Y);
                myPlayerFArm.SetPosition(myPlayer.getX() + (myPlayer.getW() - ARM_ROTATION_CENTER_X), myPlayer.getY() + ARM_ROTATION_CENTER_Y);
        }

        if (!quitLevel)
        {
                // Camera scrolling
                handleCameras(myPlayer.parentRoom->bg.getW());
        }

        // Animate player
        if (myPlayer.xMov != 0.f)
        {
                myPlayer.animation.animate(myPlayer);
                myPlayerBArm.animation.animate(myPlayerBArm);
                myPlayerFArm.animation.animate(myPlayerFArm);
        }
        else
        {
                // Stand still
                myPlayer.animation.setFrame(1, myPlayer);
                myPlayerBArm.animation.setFrame(1, myPlayer);
                myPlayerFArm.animation.setFrame(1, myPlayer);
        }

        // DRAW WORLD

        // Backgrounds (except doorframe slices, if there are any)
        for (unsigned int i = 0; i < myPlayer.parentRoom->bg.objects.size(); i++)
        {
                if (!(myPlayer.parentRoom->hasLeftDoor && i == 0) && !(myPlayer.parentRoom->hasRightDoor && i == myPlayer.parentRoom->bg.objects.size() - 1))
                {
                        window.Draw(*myPlayer.parentRoom->bg.getSprite(i));
                }
        }

        if (networkMode == MULTIPLAYER)
        {
                window.Draw(theirPlayer);
        }
                
        window.Draw(myPlayerBArm);
        window.Draw(myPlayer);
        window.Draw(myPlayerFArm);

        // Draw monsters
        for (unsigned int i = 0; i < monsters.size(); i++)
        {
                if (monsters[i] != NULL)
                {
                        // Don't draw any monsters that aren't in our room
                        if (myPlayer.parentRoom == monsters[i]->parentRoom)
                        {
                                window.Draw(*monsters[i]);
                        }
                }
        }

        // If we're in the portal room, draw it
        if (myPlayer.parentRoom == portal.parentRoom)
        {
                window.Draw(portal);
        }

        if (myPlayer.parentRoom->hasLeftDoor)
        {
                // Draw left doorframe on top of player
                window.Draw(*myPlayer.parentRoom->bg.getSprite(0));
        }
        if (myPlayer.parentRoom->hasRightDoor)
        {
                // Draw right doorframe on top of player
                window.Draw(*myPlayer.parentRoom->bg.getSprite(myPlayer.parentRoom->bg.objects.size() - 1));
        }

        if (quitLevel)
        {
                resetCamera();
        }
}

void Game::handleCameras(float levelWidth)
{
        camera.SetCenter(myPlayer.getCenterX(), getScreenHeight() / 2.f);

        // Keep cameras in the level
        if (camera.GetRect().Left < 0.f)
        {
                camera.SetCenter(getScreenWidth() / 2.f, getScreenHeight() / 2.f);
        }

        // Right side of the level
        if (camera.GetRect().Right > levelWidth)
        {
                camera.SetCenter(levelWidth - (getScreenWidth() / 2.f), getScreenHeight() / 2.f);
        }
}

void Game::handlePlayerControls(int directionFacing)
{
        // Running right
        if (window.GetInput().IsKeyDown(sf::Key::D))
        {
                // Trivial case: facing right, moving right
                if (directionFacing == FACING_RIGHT)
                {
                        myPlayer.xMov = RUN_SPEED;

                        // Facing right
                        myPlayer.FlipX(false);
                        myPlayerBArm.FlipX(false);
                        myPlayerFArm.FlipX(false);

                        // Walking forward animation
                        myPlayer.animation.setPosDirection();
                        myPlayerBArm.animation.setPosDirection();
                        myPlayerFArm.animation.setPosDirection();
                        myPlayer.animation.frameDelay = PLAYER_ANIMATION_DELAY;
                        myPlayerBArm.animation.frameDelay = PLAYER_ANIMATION_DELAY;
                        myPlayerFArm.animation.frameDelay = PLAYER_ANIMATION_DELAY;

                        myPlayer.setState(FACING_RIGHT);

                        // Did we change state? Then fix the rotation
                        if (myPlayer.getChangedState())
                        {
                                myPlayerBArm.SetRotation(-myPlayerBArm.GetRotation());
                                myPlayerFArm.SetRotation(-myPlayerFArm.GetRotation());
                        }
                }
                // We're moving right but facing left
                else if (directionFacing == FACING_LEFT)
                {
                        // Slower walk
                        myPlayer.xMov = BACKWARD_SPEED; 

                        // Facing left, so we need to flip the sprite
                        myPlayer.FlipX(true);
                        myPlayerBArm.FlipX(true);
                        myPlayerFArm.FlipX(true);

                        myPlayer.setState(FACING_LEFT);

                        // Did we change state? Then fix the rotation
                        if (myPlayer.getChangedState())
                        {
                                myPlayerBArm.SetRotation(-myPlayerBArm.GetRotation());
                                myPlayerFArm.SetRotation(-myPlayerFArm.GetRotation());
                        }

                        // Animation should reverse and slow down
                        myPlayer.animation.setNegDirection();
                        myPlayerBArm.animation.setNegDirection();
                        myPlayerFArm.animation.setNegDirection();
                        myPlayer.animation.frameDelay = PLAYER_ANIMATION_DELAY_SLOW;
                        myPlayerBArm.animation.frameDelay = PLAYER_ANIMATION_DELAY_SLOW;
                        myPlayerFArm.animation.frameDelay = PLAYER_ANIMATION_DELAY_SLOW;
                }
        }
        // Running left
        else if (window.GetInput().IsKeyDown(sf::Key::A))
        {
                // Trivial case: moving left, facing left
                if (directionFacing == FACING_LEFT)
                {
                        myPlayer.xMov = -RUN_SPEED;

                        // If moving left, we need to flip the sprite
                        myPlayer.FlipX(true);
                        myPlayerBArm.FlipX(true);
                        myPlayerFArm.FlipX(true);

                        myPlayer.animation.setPosDirection();
                        myPlayerBArm.animation.setPosDirection();
                        myPlayerFArm.animation.setPosDirection();
                        myPlayer.animation.frameDelay = PLAYER_ANIMATION_DELAY;
                        myPlayerBArm.animation.frameDelay = PLAYER_ANIMATION_DELAY;
                        myPlayerFArm.animation.frameDelay = PLAYER_ANIMATION_DELAY;

                        myPlayer.setState(FACING_LEFT);

                        // Did we change state? Then fix the rotation
                        if (myPlayer.getChangedState())
                        {
                                myPlayerBArm.SetRotation(-myPlayerBArm.GetRotation());
                                myPlayerFArm.SetRotation(-myPlayerFArm.GetRotation());
                        }
                }
                // We're moving left but facing right
                else if (directionFacing == FACING_RIGHT)
                {
                        // Slower walk
                        myPlayer.xMov = -BACKWARD_SPEED; 

                        // Face to the right
                        myPlayer.FlipX(false);
                        myPlayerBArm.FlipX(false);
                        myPlayerFArm.FlipX(false);

                        myPlayer.setState(FACING_RIGHT);

                        // Did we change state? Then fix the rotation
                        if (myPlayer.getChangedState())
                        {
                                myPlayerBArm.SetRotation(-myPlayerBArm.GetRotation());
                                myPlayerFArm.SetRotation(-myPlayerFArm.GetRotation());
                        }

                        // Animation should reverse
                        myPlayer.animation.setNegDirection();
                        myPlayerBArm.animation.setNegDirection();
                        myPlayerFArm.animation.setNegDirection();
                        myPlayer.animation.frameDelay = PLAYER_ANIMATION_DELAY_SLOW;
                        myPlayerBArm.animation.frameDelay = PLAYER_ANIMATION_DELAY_SLOW;
                        myPlayerFArm.animation.frameDelay = PLAYER_ANIMATION_DELAY_SLOW;
                }
        }
        else
        {
                myPlayer.xMov = 0;

                if (directionFacing == LEFT)
                {
                        // Facing right
                        myPlayer.FlipX(false);
                        myPlayerBArm.FlipX(false);
                        myPlayerFArm.FlipX(false);

//                        // Walking forward animation
//                        myPlayer.animation.setPosDirection();
//                        myPlayerBArm.animation.setPosDirection();
//                        myPlayerFArm.animation.setPosDirection();
//                        myPlayer.animation.frameDelay = PLAYER_ANIMATION_DELAY;
//                        myPlayerBArm.animation.frameDelay = PLAYER_ANIMATION_DELAY;
//                        myPlayerFArm.animation.frameDelay = PLAYER_ANIMATION_DELAY;

                        myPlayer.setState(FACING_RIGHT);

                        // Did we change state? Then fix the rotation
                        if (myPlayer.getChangedState())
                        {
                                myPlayerBArm.SetRotation(-myPlayerBArm.GetRotation());
                                myPlayerFArm.SetRotation(-myPlayerFArm.GetRotation());
                        }
                }
                else if (directionFacing == RIGHT)
                {
                        // If moving left, we need to flip the sprite
                        myPlayer.FlipX(true);
                        myPlayerBArm.FlipX(true);
                        myPlayerFArm.FlipX(true);

//                        myPlayer.animation.setPosDirection();
//                        myPlayerBArm.animation.setPosDirection();
//                        myPlayerFArm.animation.setPosDirection();
//                        myPlayer.animation.frameDelay = PLAYER_ANIMATION_DELAY;
//                        myPlayerBArm.animation.frameDelay = PLAYER_ANIMATION_DELAY;
//                        myPlayerFArm.animation.frameDelay = PLAYER_ANIMATION_DELAY;

                        myPlayer.setState(FACING_LEFT);

                        // Did we change state? Then fix the rotation
                        if (myPlayer.getChangedState())
                        {
                                myPlayerBArm.SetRotation(-myPlayerBArm.GetRotation());
                                myPlayerFArm.SetRotation(-myPlayerFArm.GetRotation());
                        }
                }
        }
}

void Game::resetCamera()
{
        // Set main camera and its parameters
        sf::FloatRect defaultView(0.f, 0.f, getScreenWidth(), getScreenHeight());
        camera.SetFromRect(defaultView);
        window.SetView(camera);
}

void Game::initializeMap()
{
        // Default boundaries for doorless rooms
        static const float DEFAULT_LEFT_BOUNDARY = 30.f;
        static const float DEFAULT_RIGHT_BOUNDARY = 30.f;

        // All of the rooms in the map
        rooms["Alchemy Lab"] = new Room;
        rooms["Ballroom Bar"] = new Room;
        rooms["Ballroom"] = new Room;
        rooms["Basement"] = new Room;
        rooms["Bathroom"] = new Room;
        rooms["Bedroom"] = new Room;
        rooms["Hallway 0f"] = new Room;
        rooms["Hallway 1f"] = new Room;
        rooms["Hallway 2f"] = new Room;
        rooms["Monster Brain"] = new Room;
        rooms["Monster Face"] = new Room;
        rooms["Void Hallway"] = new Room;
        rooms["Roof"] = new Room;
        rooms["Stairwell 1f"] = new Room;
        rooms["Stairwell 2f"] = new Room;
        rooms["Stairwell 3f"] = new Room;
        rooms["Study"] = new Room;

        rooms["Ballroom Bar"]->links[LEFT] = rooms["Ballroom"];
        rooms["Ballroom Bar"]->links[RIGHT] = rooms["Void Hallway"];
        rooms["Ballroom Bar"]->filenames.push_back("1f_ballroombar_01.png");
        rooms["Ballroom Bar"]->filenames.push_back("1f_ballroombar_02.png");
        rooms["Ballroom Bar"]->filenames.push_back("1f_ballroombar_03.png");
        rooms["Ballroom Bar"]->filenames.push_back("1f_ballroombar_04.png");
        rooms["Ballroom Bar"]->hasLeftDoor = true;
        rooms["Ballroom Bar"]->hasRightDoor = true;

        rooms["Void Hallway"]->links[LEFT] = rooms["Ballroom Bar"];
        rooms["Void Hallway"]->filenames.push_back("1f_voidhallway_01.png");
        rooms["Void Hallway"]->filenames.push_back("1f_voidhallway_02.png");
        rooms["Void Hallway"]->filenames.push_back("1f_voidhallway_03.png");
        rooms["Void Hallway"]->filenames.push_back("1f_voidhallway_04.png");
        rooms["Void Hallway"]->filenames.push_back("1f_voidhallway_05.png");
        rooms["Void Hallway"]->hasLeftDoor = true;
        rooms["Void Hallway"]->hasRightWall = true;

        rooms["Ballroom"]->links[LEFT] = rooms["Stairwell 1f"];
        rooms["Ballroom"]->links[RIGHT] = rooms["Ballroom Bar"];
        rooms["Ballroom"]->filenames.push_back("1f_ballroom_01.png");
        rooms["Ballroom"]->filenames.push_back("1f_ballroom_02.png");
        rooms["Ballroom"]->filenames.push_back("1f_ballroom_03.png");
        rooms["Ballroom"]->filenames.push_back("1f_ballroom_04.png");
        rooms["Ballroom"]->filenames.push_back("1f_ballroom_05.png");
        rooms["Ballroom"]->hasLeftDoor = true;
        rooms["Ballroom"]->hasRightDoor = true;

        rooms["Hallway 1f"]->links[RIGHT] = rooms["Stairwell 1f"];
        rooms["Hallway 1f"]->links[LEFT] = rooms["Study"];
        rooms["Hallway 1f"]->links[DOWN] = rooms["Basement"];
        rooms["Hallway 1f"]->filenames.push_back("1f_candlehallway_01.png");
        rooms["Hallway 1f"]->filenames.push_back("1f_candlehallway_02.png");
        rooms["Hallway 1f"]->filenames.push_back("1f_candlehallway_03.png");
        rooms["Hallway 1f"]->filenames.push_back("1f_candlehallway_04.png");
        rooms["Hallway 1f"]->hasLeftDoor = true;
        rooms["Hallway 1f"]->hasRightDoor = true;
        rooms["Hallway 1f"]->stairsDown.Left = 492;
        rooms["Hallway 1f"]->stairsDown.Right = 814;

        rooms["Basement"]->links[UP] = rooms["Hallway 1f"];
        rooms["Basement"]->links[LEFT] = rooms["Hallway 0f"];
        rooms["Basement"]->filenames.push_back("0f_basement_01.png");
        rooms["Basement"]->filenames.push_back("0f_basement_02.png");
        rooms["Basement"]->filenames.push_back("0f_basement_03.png");
        rooms["Basement"]->filenames.push_back("0f_basement_04.png");
        rooms["Basement"]->hasRightWall = true;
        rooms["Basement"]->hasLeftDoor = true;
        rooms["Basement"]->stairsUp.Left = 218;
        rooms["Basement"]->stairsUp.Right = 549;

        rooms["Hallway 0f"]->links[RIGHT] = rooms["Basement"];
        rooms["Hallway 0f"]->links[LEFT] = rooms["Alchemy Lab"];
        rooms["Hallway 0f"]->filenames.push_back("0f_candlehallway_01.png");
        rooms["Hallway 0f"]->filenames.push_back("0f_candlehallway_02.png");
        rooms["Hallway 0f"]->filenames.push_back("0f_candlehallway_03.png");
        rooms["Hallway 0f"]->filenames.push_back("0f_candlehallway_04.png");
        rooms["Hallway 0f"]->filenames.push_back("0f_candlehallway_05.png");
        rooms["Hallway 0f"]->hasRightDoor = true;
        rooms["Hallway 0f"]->hasLeftDoor = true;
        
        rooms["Study"]->links[RIGHT] = rooms["Hallway 1f"];
        rooms["Study"]->links[LEFT] = rooms["Alchemy Lab"];
        rooms["Study"]->filenames.push_back("1f_study_01.png");
        rooms["Study"]->filenames.push_back("1f_study_02.png");
        rooms["Study"]->filenames.push_back("1f_study_03.png");
        rooms["Study"]->filenames.push_back("1f_study_04.png");
        rooms["Study"]->filenames.push_back("1f_study_05.png");
        rooms["Study"]->hasRightDoor = true;
        rooms["Study"]->hasLeftDoor = true;

        rooms["Alchemy Lab"]->links[RIGHT] = rooms["Study"];
        rooms["Alchemy Lab"]->filenames.push_back("1f_alchemylab_01.png");
        rooms["Alchemy Lab"]->filenames.push_back("1f_alchemylab_02.png");
        rooms["Alchemy Lab"]->filenames.push_back("1f_alchemylab_03.png");
        rooms["Alchemy Lab"]->hasRightDoor = true;
        rooms["Alchemy Lab"]->hasLeftWall = true;

        rooms["Stairwell 1f"]->links[UP] = rooms["Stairwell 2f"];
        rooms["Stairwell 1f"]->links[LEFT] = rooms["Hallway 1f"];
        rooms["Stairwell 1f"]->links[RIGHT] = rooms["Ballroom"];
        rooms["Stairwell 1f"]->filenames.push_back("1f_stairwell_01.png");
        rooms["Stairwell 1f"]->filenames.push_back("1f_stairwell_02.png");
        rooms["Stairwell 1f"]->filenames.push_back("1f_stairwell_03.png");
        rooms["Stairwell 1f"]->filenames.push_back("1f_stairwell_04.png");
        rooms["Stairwell 1f"]->hasLeftDoor = true;
        rooms["Stairwell 1f"]->hasRightDoor = true;
        rooms["Stairwell 1f"]->stairsUp.Left = 578;
        rooms["Stairwell 1f"]->stairsUp.Right = 898;

        rooms["Stairwell 2f"]->links[LEFT] = rooms["Hallway 2f"];
        rooms["Stairwell 2f"]->links[UP] = rooms["Stairwell 3f"];
        rooms["Stairwell 2f"]->links[DOWN] = rooms["Stairwell 1f"];
        rooms["Stairwell 2f"]->filenames.push_back("2f_stairwell_01.png");
        rooms["Stairwell 2f"]->filenames.push_back("2f_stairwell_02.png");
        rooms["Stairwell 2f"]->filenames.push_back("2f_stairwell_03.png");
        rooms["Stairwell 2f"]->hasLeftDoor = true;
        rooms["Stairwell 2f"]->hasRightWall = true;
        rooms["Stairwell 2f"]->stairsDown.Left = 61;
        rooms["Stairwell 2f"]->stairsDown.Right = 405;
        rooms["Stairwell 2f"]->stairsUp.Left = 733;
        rooms["Stairwell 2f"]->stairsUp.Right = 1011;

        rooms["Hallway 2f"]->links[RIGHT] = rooms["Stairwell 2f"];
        rooms["Hallway 2f"]->links[LEFT] = rooms["Bedroom"];
        rooms["Hallway 2f"]->filenames.push_back("2f_candlehallway_01.png");
        rooms["Hallway 2f"]->filenames.push_back("2f_candlehallway_02.png");
        rooms["Hallway 2f"]->filenames.push_back("2f_candlehallway_03.png");
        rooms["Hallway 2f"]->filenames.push_back("2f_candlehallway_04.png");
        rooms["Hallway 2f"]->hasLeftDoor = true;
        rooms["Hallway 2f"]->hasRightDoor = true;

        rooms["Bedroom"]->links[RIGHT] = rooms["Hallway 2f"];
        rooms["Bedroom"]->links[LEFT] = rooms["Bathroom"];
        rooms["Bedroom"]->filenames.push_back("2f_bedroom_01.png");
        rooms["Bedroom"]->filenames.push_back("2f_bedroom_02.png");
        rooms["Bedroom"]->filenames.push_back("2f_bedroom_03.png");
        rooms["Bedroom"]->filenames.push_back("2f_bedroom_04.png");
        rooms["Bedroom"]->hasLeftDoor = true;
        rooms["Bedroom"]->hasRightDoor = true;

        rooms["Bathroom"]->links[RIGHT] = rooms["Bedroom"];
        rooms["Bathroom"]->filenames.push_back("2f_bathroom_01.png");
        rooms["Bathroom"]->filenames.push_back("2f_bathroom_02.png");
        rooms["Bathroom"]->filenames.push_back("2f_bathroom_03.png");
        rooms["Bathroom"]->hasRightDoor = true;
        rooms["Bathroom"]->hasLeftWall = true;

        rooms["Stairwell 3f"]->links[DOWN] = rooms["Stairwell 2f"];
        rooms["Stairwell 3f"]->links[UP] = rooms["Roof"];
        rooms["Stairwell 3f"]->filenames.push_back("3f_stairwell_01.png");
        rooms["Stairwell 3f"]->filenames.push_back("3f_stairwell_02.png");
        rooms["Stairwell 3f"]->leftBoundary = DEFAULT_LEFT_BOUNDARY;
        rooms["Stairwell 3f"]->rightBoundary = DEFAULT_RIGHT_BOUNDARY;
        rooms["Stairwell 3f"]->hasLeftWall = true;
        rooms["Stairwell 3f"]->hasRightWall = true;
        rooms["Stairwell 3f"]->stairsDown.Left = 0;
        rooms["Stairwell 3f"]->stairsDown.Right = 405;
        rooms["Stairwell 3f"]->stairsUp.Left = 733;
        rooms["Stairwell 3f"]->stairsUp.Right = 1011;

        rooms["Roof"]->links[DOWN] = rooms["Stairwell 3f"];
        rooms["Roof"]->links[RIGHT] = rooms["Monster Face"];
        rooms["Roof"]->filenames.push_back("4f_roof_01.png");
        rooms["Roof"]->filenames.push_back("4f_roof_02.png");
        rooms["Roof"]->filenames.push_back("4f_roof_03.png");
        rooms["Roof"]->filenames.push_back("4f_roof_04.png");
        rooms["Roof"]->filenames.push_back("4f_roof_05.png");
        rooms["Roof"]->hasLeftWall = true;
        rooms["Roof"]->stairsDown.Left = 740;
        rooms["Roof"]->stairsDown.Right = 1371;

        rooms["Monster Face"]->links[LEFT] = rooms["Roof"];
        rooms["Monster Face"]->links[RIGHT] = rooms["Monster Brain"];
        rooms["Monster Face"]->filenames.push_back("4f_monsterface_01.png");
        rooms["Monster Face"]->filenames.push_back("4f_monsterface_02.png");
        rooms["Monster Face"]->filenames.push_back("4f_monsterface_03.png");
        rooms["Monster Face"]->hasRightDoor = true;

        rooms["Monster Brain"]->links[LEFT] = rooms["Monster Face"];
        rooms["Monster Brain"]->filenames.push_back("4f_monsterbrain_01.png");
        rooms["Monster Brain"]->filenames.push_back("4f_monsterbrain_02.png");
        rooms["Monster Brain"]->hasRightWall = true;

        myPlayer.parentRoom = rooms["Stairwell 1f"];
        theirPlayer.parentRoom = rooms["Stairwell 1f"];

        // Load each map's image
        std::map<std::string, Room*>::iterator it = rooms.begin();
        for (it; it != rooms.end(); it++)
        {
                if (!(it->second->loadBackground()))
                {
                        // Couldn't load background!
                        quitLevel = true;
                        writeError("hostGame: Couldn't load a background");
                }
        }
}

void Game::deleteMap()
{
        if (!rooms.empty())
        {
                std::map<std::string, Room*>::iterator it = rooms.begin();

                for (it; it != rooms.end(); it++)
                {
                        // Free memory allocated for this room
                        delete it->second;
                }

                // Clear the keys out
                rooms.clear();

                myPlayer.parentRoom = NULL;
                theirPlayer.parentRoom = NULL;
        }
}

float Game::distance(float x1, float y1, float x2, float y2)
{
        return std::sqrt(std::pow(x2 - x1, 2) + std::pow(y2 - y1, 2));
}

void Game::spawnMonsters(bool initializedMonsters)
{
        // Monsters to scatter throughout the mansion before starting portal spawning
        static const int INITIAL_MONSTER_AMOUNT = 5;

        // Absolute maximum number of allowed monsters
        static const int MAX_MONSTERS = 25;

        if (!initializedMonsters)
        {
                sf::Image *_monster = new sf::Image;

                if (!_monster->LoadFromFile("./rc/game/monster_vacuum.png"))
                {
                        writeError("Couldn't load monster_vacuum.png");
                        quitLevel = true;
                }
                _monster->SetSmooth(false);

                // Set the image for all monsters
                Monster::image = _monster;

                for (int i = 0; i < INITIAL_MONSTER_AMOUNT; i++)
                {
                        livingMonsters++;

                        // Add a monster
                        Monster *m = new Monster;

                        m->SetImage(*m->image);

                        // Pick a random room
                        std::map<std::string, Room*>::iterator it;

                        // Repeat until we have a random room that no player is in
                        do
                        {
                                it = rooms.begin();

                                int randRoom = sf::Randomizer::Random(0, rooms.size() - 1);

                                // Iterate through the map of rooms until we have the randomly selected room
                                for (int i = 0; i < randRoom && it != rooms.end(); i++)
                                { 
                                        it++;
                                }
                        } while (it->second == myPlayer.parentRoom);

                        // Move the monster into a random room
                        m->parentRoom = it->second;

                        m->SetSubRect(sf::IntRect(0, 0, MONSTER_FRAME_WIDTH, MONSTER_FRAME_HEIGHT));

                        // Animation specs
                        m->animation.setRows(MONSTER_ROWS, m->GetImage());
                        m->animation.setColumns(MONSTER_COLUMNS, m->GetImage());
                        m->animation.setFirstFrame(MONSTER_FIRST_FRAME);
                        m->animation.setLastFrame(MONSTER_LAST_FRAME);

                        m->setState(WANDERING);

                        // Set random starting position in the room
                        m->SetPosition(sf::Randomizer::Random(0.f, m->parentRoom->bg.getW() - m->getW()), FLOOR_HEIGHT - m->getH());

                        // Set random movement direction
                        int dir = 1;
                        do
                        {
                                dir = sf::Randomizer::Random(-1, 1);
                        } while (dir == 0);
                        m->xMov *= float(dir);

                        // Find the first non-NULL spot to put the monster in, or add it to the end
                        bool foundSlot = false;
                        for (unsigned int i = 0; i < monsters.size() && !foundSlot; i++)
                        {
                                if (monsters[i] == NULL)
                                {
                                        monsters[i] = m;
                                        foundSlot = true;
                                }
                        }

                        if (!foundSlot)
                                monsters.push_back(m);
                }

                initializedMonsters = true;
        }

        // Time to spawn a new monster yet?
        if (livingMonsters < MAX_MONSTERS && portal.spawnTimer.GetElapsedTime() > portal.spawnDelay)
        {
                livingMonsters++;

                // Add a monster
                Monster *m = new Monster;

                m->SetImage(*m->image);

                // Spawn the monster in the portal room
                m->parentRoom = portal.parentRoom;

                m->SetSubRect(sf::IntRect(0, 0, MONSTER_FRAME_WIDTH, MONSTER_FRAME_HEIGHT));

                // Animation specs
                m->animation.setRows(MONSTER_ROWS, m->GetImage());
                m->animation.setColumns(MONSTER_COLUMNS, m->GetImage());
                m->animation.setFirstFrame(MONSTER_FIRST_FRAME);
                m->animation.setLastFrame(MONSTER_LAST_FRAME);

                m->setState(WANDERING);

                // Spawn inside the portal
                m->SetPosition(portal.getX() - (m->getW() / 2.f), FLOOR_HEIGHT - m->getH());

                // Set random movement direction
                int dir = 1;
                do
                {
                        dir = sf::Randomizer::Random(-1, 1);
                } while (dir == 0);
                m->xMov *= float(dir);

                // Can we find an empty spot to put this new monster in?
                bool foundSlot = false;
                for (unsigned int i = 0; i < monsters.size() && !foundSlot; i++)
                {
                        // Add it to an empty slot
                        if (monsters[i] == NULL)
                        {
                                monsters[i] = m;
                                foundSlot = true;
                        }
                }

                if (!foundSlot)
                {
                        monsters.push_back(m);
                }

                portal.spawnTimer.Reset();
        }
}

void Game::deleteMonsters()
{
        for (unsigned int i = 0; i < monsters.size(); i++)
        {
                if (monsters[i] != NULL)
                {
                        delete monsters[i];
                }
                monsters[i] = NULL;
        }

        // Free up the image associated with the monsters
        if (Monster::image != NULL)
        {
                delete Monster::image;
        }
}

void Game::deletePickups()
{
        for (unsigned int i = 0; i < pickups.size(); i++)
        {
                delete pickups[i];
                pickups[i] = NULL;
        }

        delete Pickup::image;
}

void Game::moveAndAnimateMonsters(float deltaTime)
{
        for (unsigned int i = 0; i < monsters.size(); i++)
        {
                if (monsters[i] != NULL)
                {
                        monsters[i]->move(deltaTime);

                        monsters[i]->animation.animate(*monsters[i]);
                }
        }
}

void Game::monsterAI()
{
        for (unsigned int i = 0; i < monsters.size(); i++)
        {
                if (monsters[i] != NULL)
                {
                        Room *room = monsters[i]->parentRoom;

                        // If we're hunting, we're in the target room, and the player is STILL in this room
                        if (monsters[i]->getState() == HUNTING && room == myPlayer.parentRoom)
                        {
                                // Go to the nearest player

                                // Turn to face the player and move toward them
                                if (myPlayer.getX() < monsters[i]->getX() && monsters[i]->xMov > 0)
                                {
                                        monsters[i]->xMov *= -1;
                                }
                                else if (myPlayer.getX() > monsters[i]->getX() && monsters[i]->xMov < 0)
                                {
                                        monsters[i]->xMov *= -1;
                                }
                        }

                        // COLLISION DETECTION

                        // Check if the monster is colliding with the walls
                        if (room->hasLeftWall && monsters[i]->getX() < room->leftBoundary)
                        {
                                monsters[i]->SetPosition(room->leftBoundary, monsters[i]->getY());

                                // Reverse direction
                                monsters[i]->xMov = -monsters[i]->xMov;
                        }
                        else if (room->hasRightWall && monsters[i]->getX() + monsters[i]->getW() > room->bg.getW() - room->rightBoundary)
                        {
                                monsters[i]->SetPosition(room->bg.getW() - room->rightBoundary - monsters[i]->getW(), monsters[i]->getY());

                                // Reverse direction
                                monsters[i]->xMov = -monsters[i]->xMov;
                        }

                        // Monster leaving right side of room?
                        if (monsters[i]->getX() > room->bg.getW())
                        {
                                // If there's a connected room on the right
                                if (room->links[RIGHT] != NULL)
                                {
                                        // Move into the left side of that room
                                        room = room->links[RIGHT];

                                        monsters[i]->SetPosition(-monsters[i]->getW() / 2.f, monsters[i]->getY());
                                }
                        }
                        // Monster leaving left side of room?
                        else if (monsters[i]->getX() < -monsters[i]->getW())
                        {
                                // If there's a connected room on the left
                                if (room->links[LEFT] != NULL)
                                {
                                        // Move into the left side of that room
                                        room = room->links[LEFT];

                                        monsters[i]->SetPosition(room->bg.getW() - (monsters[i]->getW() / 2.f), monsters[i]->getY());
                                }
                        }

                        // HUNT FOR PLAYERS

                        // Is there a player in this room?
                        if (room == myPlayer.parentRoom)
                        {
                                // Already in the target room
                                monsters[i]->targetRoom = room;

                                // Go into hunt mode
                                monsters[i]->setState(HUNTING);
                        }
                        // Does the left room have a player?
                        else if (room->links[LEFT] == myPlayer.parentRoom)
                        {
                                // Start seeking this room
                                monsters[i]->targetRoom = room->links[LEFT];

                                // Move to the left
                                if (monsters[i]->xMov > 0) monsters[i]->xMov *= -1;

                                // Go into hunt mode
                                monsters[i]->setState(HUNTING);
                        }
                        // Does the right room have a player?
                        else if (room->links[RIGHT] == myPlayer.parentRoom)
                        {
                                // Start seeking this room
                                monsters[i]->targetRoom = room->links[RIGHT];

                                // Move to the right
                                if (monsters[i]->xMov < 0) monsters[i]->xMov *= -1;

                                // Go into hunt mode
                                monsters[i]->setState(HUNTING);
                        }
        //                // Does the top room have a player?
        //                else if (room->links[UP] == myPlayer.parentRoom)
        //                {
        //                        // Start seeking this room
        //                        monsters[i]->targetRoom = room->links[UP];

        //                        // Go into hunt mode
        //                        monsters[i]->setState(HUNTING);

        //                        // If we're in the doorway, go up
        //                        // Else move to the doorway
        //                }
        //                // Does the bottom room have a player?
        //                else if (room->links[DOWN] == myPlayer.parentRoom)
        //                {
        //                        // Start seeking this room
        //                        monsters[i]->targetRoom = room->links[DOWN];

        //                        // Go into hunt mode
        //                        monsters[i]->setState(HUNTING);

        //                        // If we're in the doorway, go down
        //                        // Else move to the doorway
        //                }
                        else
                        {
                                // Adjacent rooms are empty, so continue to wander
                                monsters[i]->setState(WANDERING);
                        }

                        // Update room changes
                        monsters[i]->parentRoom = room;

                        // Monster's color should reflect their health
                        float scaleFactor = (255.f / monsters[i]->maxHealth);
                        // Approaches 0 as health approaches MAX_HEALTH
                        sf::Uint8 colorDrain = 255 - ((monsters[i]->maxHealth - monsters[i]->getHealth()) * (255 / monsters[i]->maxHealth));
                        monsters[i]->SetColor(sf::Color(255, colorDrain, colorDrain));
                }
        }
}

int max(int a, int b)
{
        if (a >= b) return a;
        else return b;
}

int min(int a, int b)
{
        if (a <= b) return a;
        else return b;
}

bool inVector(int i, std::vector<int>& v)
{
        for (unsigned int i = 0; i < v.size(); i++)
        {
                if (v[i] == i)
                {
                        return true;
                }
        }

        return false;
}
